import torch
from torch import nn, Tensor

from analogvnn.backward.BackwardIdentity import BackwardIdentity
from analogvnn.fn.reduce_precision import reduce_precision
from analogvnn.nn.module.Layer import Layer
from analogvnn.utils.common_types import TENSOR_OPERABLE

__all__ = ['ReducePrecision']


class ReducePrecision(Layer, BackwardIdentity):
    """Implements the reduce precision function.

    Attributes:
        precision (nn.Parameter): the precision of the output tensor.
        divide (nn.Parameter): the rounding value that is if divide is 0.5,
         then 0.6 will be rounded to 1.0 and 0.4 will be rounded to 0.0.
    """
    __constants__ = ['precision', 'divide']
    precision: nn.Parameter
    divide: nn.Parameter

    def __init__(self, precision: int = None, divide: float = 0.5):
        """initialize the reduce precision function.

        Args:
            precision (int): the precision of the output tensor.
            divide (float): the rounding value that is if divide is 0.5,
             then 0.6 will be rounded to 1.0 and 0.4 will be rounded to 0.0.
        """
        super(ReducePrecision, self).__init__()
        if precision < 1:
            raise ValueError(f"precision has to be more than 0, but got {precision}")

        if precision != int(precision):
            raise ValueError(f"precision must be int, but got {precision}")

        if not (0 <= divide <= 1):
            raise ValueError(f"divide must be between 0 and 1, but got {divide}")

        self.precision = nn.Parameter(torch.tensor(precision), requires_grad=False)
        self.divide = nn.Parameter(torch.tensor(divide), requires_grad=False)

    @property
    def precision_width(self) -> Tensor:
        """the precision width

        Returns:
            Tensor: the precision width
        """
        return 1 / self.precision

    @property
    def bit_precision(self) -> Tensor:
        """the bit precision of the ReducePrecision module.

        Returns:
            Tensor: the bit precision of the ReducePrecision module.
        """
        return torch.log2(self.precision + 1)

    @staticmethod
    def convert_to_precision(bit_precision: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """convert the bit precision to the precision.

        Args:
            bit_precision (TENSOR_OPERABLE): the bit precision.

        Returns:
            TENSOR_OPERABLE: the precision.
        """
        return 2 ** bit_precision - 1

    def extra_repr(self) -> str:
        """the extra __repr__ string of the ReducePrecision module.

        Returns:
            str: string
        """
        return f'precision={int(self.precision)}, divide={float(self.divide):0.2f}'

    def forward(self, x: Tensor) -> Tensor:
        """forward function of the ReducePrecision module.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return reduce_precision(x, self.precision, self.divide)



import sys

import importlib_metadata

__package__ = "analogvnn"
__version__ = importlib_metadata.version(__package__)
__author__ = "Vivswan Shah (vivswanshah@pitt.edu)"

if sys.version_info < (3, 7, 0):
    import warnings

    warnings.warn(
        'The installed Python version reached its end-of-life. Please upgrade to a newer Python version for receiving '
        'further gdshelpers updates.', Warning)



from __future__ import annotations

from typing import Tuple

import torch

__all__ = ['CPUCuda', 'is_cpu_cuda']


class CPUCuda:
    """CPUCuda is a class that can be used to get, check and set the device.

    Attributes:
        _device (torch.device): The device.
        device_name (str): The name of the device.
    """
    _device: torch.device
    device_name: str

    def __init__(self):
        """Initialize the CPUCuda class."""
        self._device = None
        self.device_name = None
        self.reset_device()

    def reset_device(self):
        """Reset the device to the default device.

        Returns:
            CPUCuda: self
        """
        self.set_device(f"cuda:{torch.cuda.current_device()}" if torch.cuda.is_available() else "cpu")
        return self

    def set_device(self, device_name: str) -> CPUCuda:
        """Set the device to the given device name.

        Args:
            device_name (str): the device name.

        Returns:
            CPUCuda: self
        """

        self._device = torch.device(device_name)
        self.device_name = self._device.type
        return self

    @property
    def is_cuda(self) -> bool:
        """Check if the device is cuda.

        Returns:
            bool: True if the device is cuda, False otherwise.
        """
        return "cuda" in self.device_name

    @property
    def device(self) -> torch.device:
        """Get the device.

        Returns:
            torch.device: the device.
        """
        return self._device

    @property
    def is_using_cuda(self) -> Tuple[torch.device, bool]:
        """Check if the device is cuda.

        Returns:
            tuple: the device and True if the device is cuda, False otherwise.
        """
        return self.device, self.is_cuda

    def get_module_device(self, module) -> torch.device:
        """Get the device of the given module.

        Args:
            module (torch.nn.Module): the module.

        Returns:
            torch.device: the device of the module.
        """

        # noinspection PyBroadException
        try:
            device: torch.device = getattr(module, "device", None)
            if device is None:
                device = next(module.parameters()).device
            return device
        except Exception:
            return self._device


is_cpu_cuda: CPUCuda = CPUCuda()
"""CPUCuda: The CPUCuda instance."""



from __future__ import annotations

from abc import ABC
from typing import Callable

from torch import nn, Tensor

from analogvnn.backward.BackwardModule import BackwardModule
from analogvnn.utils.common_types import TENSORS

__all__ = ['BackwardFunction']


class BackwardFunction(BackwardModule, ABC):
    """The backward module that uses a function to compute the backward gradient.

    Attributes:
        _backward_function (Callable): The function used to compute the backward gradient.
    """
    _backward_function: Callable

    def __init__(self, backward_function: Callable, layer: nn.Module = None):
        """Initializes the backward module.

        Args:
            backward_function (Callable): The function used to compute the backward gradient.
            layer (nn.Module): The layer that this backward module is associated with.
        """
        super().__init__(layer)
        self._backward_function = backward_function

    @property
    def backward_function(self) -> Callable:
        """The function used to compute the backward gradient.

        Returns:
            Callable: The function used to compute the backward gradient.
        """
        return self._backward_function

    @backward_function.setter
    def backward_function(self, backward_function: Callable):
        """Sets the function used to compute the backward gradient with.

        Args:
            backward_function (Callable): The function used to compute the backward gradient with.
        """
        self.set_backward_function(backward_function)

    def set_backward_function(self, backward_function: Callable) -> BackwardFunction:
        """Sets the function used to compute the backward gradient with.

        Args:
            backward_function (Callable): The function used to compute the backward gradient with.

        Returns:
            BackwardFunction: self.
        """
        self._backward_function = backward_function
        return self

    def backward(self, *grad_output: Tensor, **grad_output_kwarg: Tensor) -> TENSORS:
        """Computes the backward gradient of the input of the layer with respect to the output of the layer
        using the backward function.

        Args:
            *grad_output (Tensor): The gradients of the output of the layer.
            **grad_output_kwarg (Tensor): The gradients of the output of the layer.

        Returns:
            TENSORS: The gradients of the input of the layer.

        Raises:
            NotImplementedError: If the backward function is not set.
        """
        if self._backward_function is None:
            raise ValueError("set backward_function first before invoking backward")

        return self._backward_function(*grad_output, **grad_output_kwarg)



from __future__ import annotations

import uuid
from typing import Dict, Union, Callable, List, Tuple

import networkx as nx
import torch
from torch import nn

from analogvnn.backward.BackwardModule import BackwardModule
from analogvnn.graph.AccumulateGrad import AccumulateGrad
from analogvnn.graph.AcyclicDirectedGraph import AcyclicDirectedGraph
from analogvnn.graph.ArgsKwargs import ArgsKwargs, InputOutput, ArgsKwargsOutput
from analogvnn.graph.GraphEnum import GraphEnum, GRAPH_NODE_TYPE
from analogvnn.nn.module.Layer import Layer
from analogvnn.utils.common_types import TENSORS

__all__ = ['BackwardGraph']


class BackwardGraph(AcyclicDirectedGraph):
    """The backward graph."""

    def __call__(self, gradient: TENSORS = None) -> ArgsKwargsOutput:
        """Backward pass through the backward graph

        Args:
            gradient (TENSORS): gradient of the loss function w.r.t. the output of the forward graph

        Returns:
            ArgsKwargsOutput: gradient of the inputs function w.r.t. loss
        """
        self.graph_state.ready_for_backward(exception=True)

        if len(gradient) == 0:
            gradient = None
        elif len(gradient) == 1:
            gradient = gradient[0]

        if self.graph_state.use_autograd_graph:
            result = self.graph_state.loss.backward(gradient=gradient)
        else:
            grad_outputs = torch.autograd.grad(
                outputs=self.graph_state.loss,
                inputs=self.graph_state.outputs.args,
                grad_outputs=gradient,
                retain_graph=True
            )
            result = self.calculate(*grad_outputs)

        self.graph_state.set_loss(None)

        return result

    def compile(self, is_static=True):
        """Compile the graph

        Args:
            is_static (bool): If True, the graph is not changing during runtime and will be cached.

        Returns:
            BackwardGraph: self.

        Raises:
            ValueError: If no forward pass has been performed yet.
        """
        if not self.graph.has_node(self.OUTPUT):
            raise ValueError("OUTPUT doesn't exist in the forward graph. Please preform a forward pass first.")

        return super().compile(is_static=is_static)

    def from_forward(self, forward_graph: Union[AcyclicDirectedGraph, nx.DiGraph]) -> BackwardGraph:
        """Create a backward graph from inverting forward graph

        Args:
            forward_graph (Union[AcyclicDirectedGraph, nx.DiGraph]): The forward graph.

        Returns:
            BackwardGraph: self.
        """
        if isinstance(forward_graph, AcyclicDirectedGraph):
            forward_graph = forward_graph.graph

        graph = forward_graph.reverse(copy=True)
        for _, _, attr in graph.edges(data=True):
            attr["in_arg"], attr["out_arg"] = attr["out_arg"], attr["in_arg"]
            attr["in_kwarg"], attr["out_kwarg"] = attr["out_kwarg"], attr["in_kwarg"]
            attr["label"] = AcyclicDirectedGraph._create_edge_label(**attr)

        new_graph = nx.MultiDiGraph()
        for v in graph.nodes():
            if v == self.OUTPUT:
                continue
            all_predecessors = list(graph.predecessors(v))

            if len(all_predecessors) == 1 and len(graph.get_edge_data(all_predecessors[0], v)) == 1:
                attr = graph.get_edge_data(all_predecessors[0], v)[0]
                if attr["in_arg"] == attr["in_kwarg"] == attr["in_arg"] == attr["in_arg"] is True:
                    new_graph.add_edge(all_predecessors[0], v, **attr)
                    continue

            akc = AccumulateGrad(v)
            for u in all_predecessors:
                for key, attr in graph.get_edge_data(u, v).items():
                    if attr["in_arg"] is None or attr["in_kwarg"] is None:
                        uuid_str = str(uuid.uuid4()).replace("-", "")
                        new_graph.add_edge(u, akc, **{
                            "in_arg": attr["in_arg"],
                            "in_kwarg": attr["in_kwarg"],
                            "out_arg": None,
                            "out_kwarg": uuid_str,
                            "real_label": " ".join(attr["label"].split(" ")[:-1] + ["{" + uuid_str + "}"]),
                            "label": attr["label"]
                        })
                        akc.input_output_connections[uuid_str] = {
                            **attr,
                            "from": u,
                        }
                    else:
                        uuid_str = str(uuid.uuid4()).replace("-", "")
                        new_graph.add_edge(u, akc, **{
                            "in_arg": True,
                            "in_kwarg": None,
                            "out_arg": None,
                            "out_kwarg": uuid_str,
                            "real_label": "[] -> {" + uuid_str + "}",
                            "label": "[] -> []",
                        })
                        akc.input_output_connections[uuid_str] = {
                            **attr,
                            "in_kwarg": None,
                            "out_kwarg": None,
                            "from": u,
                        }
                        uuid_str = str(uuid.uuid4()).replace("-", "")
                        new_graph.add_edge(u, akc, **{
                            "in_arg": None,
                            "in_kwarg": True,
                            "out_arg": None,
                            "out_kwarg": uuid_str,
                            "real_label": "{} -> {" + uuid_str + "}",
                            "label": "{} -> {}",
                        })
                        akc.input_output_connections[uuid_str] = {
                            **attr,
                            "in_arg": None,
                            "out_arg": None,
                            "from": u,
                        }

            new_graph.add_edge(akc, v, **{
                "in_arg": True,
                "in_kwarg": True,
                "out_arg": True,
                "out_kwarg": True,
                "len": 0,
            })

        for v in graph.nodes():
            new_graph.nodes[v]["fillcolor"] = "lightblue"
        self.graph = new_graph
        return self

    @torch.no_grad()
    def calculate(self, *args, **kwargs) -> ArgsKwargsOutput:
        """Calculate the gradient of the whole graph w.r.t. loss

        Args:
            *args: The gradients args of outputs.
            **kwargs: The gradients kwargs of outputs.

        Returns:
            ArgsKwargsOutput: The gradient of the inputs function w.r.t. loss.

        Raises:
            ValueError: If no forward pass has been performed yet.
        """

        if self.graph_state.forward_input_output_graph is None:
            raise ValueError("No forward pass has been performed yet. Please preform a forward pass first.")

        input_output_graph = self._pass(self.OUTPUT, *args, **kwargs)
        self.graph_state.forward_input_output_graph = None

        if self.INPUT in input_output_graph:
            return ArgsKwargs.from_args_kwargs_object(input_output_graph[self.INPUT].outputs)
        else:
            return None

    def _pass(self, from_node: GRAPH_NODE_TYPE, *args, **kwargs) -> Dict[GRAPH_NODE_TYPE, InputOutput]:
        """Perform the backward pass through the graph

        Args:
            from_node (GRAPH_NODE_TYPE): The node to start the backward pass from.
            *args: The gradients args of outputs.
            **kwargs: The gradients kwargs of outputs.

        Returns:
            Dict[GRAPH_NODE_TYPE, InputOutput]: The input and output gradients of each node.
        """
        static_graph: List[Tuple[GRAPH_NODE_TYPE, List[GRAPH_NODE_TYPE]]] = self._create_static_sub_graph(from_node)
        input_output_graph: Dict[GRAPH_NODE_TYPE, InputOutput] = {
            from_node: InputOutput(inputs=ArgsKwargs(
                args=args,
                kwargs=kwargs
            ))
        }
        for module, predecessors in static_graph:
            if module != from_node:
                inputs = self.parse_args_kwargs(input_output_graph, module, predecessors)
                input_output_graph[module] = InputOutput(inputs=inputs)

            if isinstance(module, GraphEnum):
                input_output_graph[module].outputs = input_output_graph[module].inputs
                continue

            outputs = self._calculate_gradients(
                module,
                input_output_graph[module]
            )
            input_output_graph[module].outputs = ArgsKwargs.to_args_kwargs_object(outputs)

        return input_output_graph

    def _calculate_gradients(
            self,
            module: Union[AccumulateGrad, Layer, BackwardModule, Callable],
            grad_outputs: InputOutput
    ) -> ArgsKwargs:
        """Calculate the gradient of a module w.r.t. outputs of the module using the output's gradients.

        Args:
            module (Union[AccumulateGrad, Layer, BackwardModule, Callable]): The module to calculate the gradient of.
            grad_outputs (InputOutput): The gradients of the output of the module.

        Returns:
            ArgsKwargs: The input gradients of the module.
        """
        if module in self.graph_state.forward_input_output_graph:
            module_inputs_outputs = self.graph_state.forward_input_output_graph[module]
        else:
            module_inputs_outputs = None

        if grad_outputs.inputs.is_empty():
            return ArgsKwargs()

        if isinstance(module, AccumulateGrad):
            return module.grad(
                grad_outputs_args_kwargs=grad_outputs.inputs,
                forward_input_output_graph=self.graph_state.forward_input_output_graph,
            )

        if isinstance(module, Layer) and isinstance(module.backward_function, BackwardModule):
            module = module.backward_function

        if isinstance(module, BackwardModule):
            grad_inputs = module._call_impl_backward(*grad_outputs.inputs.args, **grad_outputs.inputs.kwargs)
            return ArgsKwargs.to_args_kwargs_object(grad_inputs)

        grad_dict = {}
        inputs = module_inputs_outputs.inputs.args + list(module_inputs_outputs.inputs.kwargs.values())
        outputs = []
        outputs_grads = []
        module_parameters = []

        if len(inputs) == 0:
            return ArgsKwargs(
                args=[torch.zeros_like(i) for i in module_inputs_outputs.inputs.args],
                kwargs={key: torch.zeros_like(value) for key, value in module_inputs_outputs.inputs.kwargs.items()}
            )

        for i in range(len(module_inputs_outputs.outputs.args)):
            outputs.append(module_inputs_outputs.outputs.args[i])
            outputs_grads.append(grad_outputs.inputs.args[i])

        for i in module_inputs_outputs.outputs.kwargs:
            outputs.append(module_inputs_outputs.outputs.kwargs[i])
            outputs_grads.append(grad_outputs.inputs.kwargs[i])

        if isinstance(module, nn.Module):
            module_parameters = list(module.parameters())
            inputs += module_parameters

        out_grads = torch.autograd.grad(
            outputs=outputs,
            inputs=inputs,
            grad_outputs=outputs_grads,
            retain_graph=True,
            allow_unused=True
        )
        for i, v in enumerate(out_grads):
            grad_dict[inputs[i]] = v

        for i in module_parameters:
            if grad_dict[i] is None:
                continue

            if i.grad is None:
                i.grad = grad_dict[i]
            else:
                i.grad += grad_dict[i]

        return ArgsKwargs(
            args=[grad_dict[i] for i in module_inputs_outputs.inputs.args],
            kwargs={key: grad_dict[value] for key, value in module_inputs_outputs.inputs.kwargs.items()}
        )



from typing import Optional

import torch
from torch import Tensor

from analogvnn.nn.normalize.Normalize import Normalize

__all__ = ['Clamp', 'Clamp01']


class Clamp(Normalize):
    """Implements the clamp normalization function with range [-1, 1].
    """

    @staticmethod
    def forward(x: Tensor):
        """forward pass of the clamp normalization function with range [-1, 1].

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return torch.clamp(x, min=-1, max=1)

    def backward(self, grad_output: Optional[Tensor]) -> Optional[Tensor]:
        """backward pass of the clamp normalization function with range [-1, 1].

        Args:
            grad_output (Optional[Tensor]): the gradient of the output tensor.

        Returns:
            Optional[Tensor]: the gradient of the input tensor.
        """
        x = self.inputs
        grad = ((-1 <= x) * (x <= 1.)).type(torch.float)
        return grad_output * grad


class Clamp01(Normalize):
    """Implements the clamp normalization function with range [0, 1].
    """

    @staticmethod
    def forward(x: Tensor):
        """forward pass of the clamp normalization function with range [0, 1].

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return torch.clamp(x, min=0, max=1)

    def backward(self, grad_output: Optional[Tensor]) -> Optional[Tensor]:
        """backward pass of the clamp normalization function with range [0, 1].

        Args:
            grad_output (Optional[Tensor]): the gradient of the output tensor.

        Returns:
            Optional[Tensor]: the gradient of the input tensor.
        """
        x = self.inputs
        grad = ((0 <= x) * (x <= 1.)).type(torch.float)
        return grad_output * grad



import math
from typing import Optional

import torch
from torch import nn, Tensor

from analogvnn.backward.BackwardModule import BackwardModule
from analogvnn.fn.to_matrix import to_matrix
from analogvnn.nn.module.Layer import Layer

__all__ = ['Linear', 'LinearBackpropagation']


class LinearBackpropagation(BackwardModule):
    """The backpropagation module of a linear layer.
    """

    def forward(self, x: Tensor):
        """Forward pass of the linear layer.

        Args:
            x (Tensor): The input of the linear layer.

        Returns:
            Tensor: The output of the linear layer.
        """
        if self.bias is not None:
            y = torch.addmm(self.bias, x, self.weight.t())
        else:
            y = torch.mm(x, self.weight.t())

        return y

    def backward(self, grad_output: Optional[Tensor], weight: Optional[Tensor] = None) -> Optional[Tensor]:
        """Backward pass of the linear layer.

        Args:
            grad_output (Optional[Tensor]): The gradient of the output.
            weight (Optional[Tensor]): The weight of the layer.

        Returns:
            Optional[Tensor]: The gradient of the input.
        """
        grad_output = to_matrix(grad_output)

        weight = to_matrix(self.weight if weight is None else weight)
        grad_input = grad_output @ weight

        self.set_grad_of(self.weight, torch.mm(grad_output.t(), self.inputs))
        self.set_grad_of(self.bias, grad_output.sum(0))
        return grad_input


class Linear(Layer):
    """A linear layer.

    Attributes:
        in_features (int): The number of input features.
        out_features (int): The number of output features.
        weight (nn.Parameter): The weight of the layer.
        bias (nn.Parameter): The bias of the layer.
    """
    __constants__ = ['in_features', 'out_features']
    in_features: int
    out_features: int
    weight: nn.Parameter
    bias: Optional[nn.Parameter]

    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        """Create a new linear layer.

        Args:
            in_features (int): The number of input features.
            out_features (int): The number of output features.
            bias (bool): True if the layer has a bias.
        """
        super(Linear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features

        self.weight = nn.Parameter(torch.Tensor(out_features, in_features))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)

        self.set_backward_function(LinearBackpropagation)
        self.reset_parameters()

    def reset_parameters(self):
        """Reset the parameters of the layer."""
        nn.init.xavier_uniform_(self.weight)
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)

    def extra_repr(self) -> str:
        """Extra representation of the linear layer.

        Returns:
            str: The extra representation of the linear layer.
        """
        return f'in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}'



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Dict, Any, Union, Optional

__all__ = ['InputOutput', 'ArgsKwargs', 'ArgsKwargsInput', 'ArgsKwargsOutput']


@dataclass
class InputOutput:
    """Inputs and outputs of a module.

    Attributes:
        inputs (Optional[ArgsKwargs]): Inputs of a module.
        outputs (Optional[ArgsKwargs]): Outputs of a module.
    """
    inputs: Optional[ArgsKwargs] = None
    outputs: Optional[ArgsKwargs] = None


@dataclass
class ArgsKwargs:
    """The arguments.

    Attributes:
        args (List): The arguments.
        kwargs (Dict): The keyword arguments.
    """
    args: List
    kwargs: Dict

    def __init__(self, args=None, kwargs=None):
        """Initialize the ArgsKwargs object.

        Args:
            args: The arguments.
            kwargs: The keyword arguments.
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        if isinstance(args, tuple):
            args = list(args)

        if not isinstance(args, List):
            args = [args]

        self.args = args
        self.kwargs = kwargs

    def is_empty(self):
        """Returns whether the ArgsKwargs object is empty."""
        return len(self.args) == 0 and not bool(self.kwargs)

    def __repr__(self):
        """Returns a string representation of the parameter."""
        return f"ArgsKwargs(args={self.args}, kwargs={self.kwargs})"

    @classmethod
    def to_args_kwargs_object(cls, outputs: ArgsKwargsInput) -> ArgsKwargs:
        """ Convert the output of a module to ArgsKwargs object

        Args:
            outputs: The output of a module

        Returns:
            ArgsKwargs: The ArgsKwargs object
        """
        if isinstance(outputs, cls):
            pass
        elif isinstance(outputs, dict):
            outputs = cls(kwargs=outputs)
        elif isinstance(outputs, tuple) and len(outputs) == 2 and isinstance(outputs[1], dict):
            outputs = cls(args=outputs[0], kwargs=outputs[1])
        else:
            outputs = cls(args=outputs)
        return outputs

    @staticmethod
    def from_args_kwargs_object(outputs: ArgsKwargs) -> ArgsKwargsOutput:
        """Convert ArgsKwargs to object or tuple or dict

        Args:
            outputs (ArgsKwargs): ArgsKwargs object

        Returns:
            ArgsKwargsOutput: object or tuple or dict
        """
        if len(outputs.kwargs.keys()) > 0:
            return outputs
        elif len(outputs.args) > 1:
            return outputs.args
        elif len(outputs.args) == 1:
            return outputs.args[0]
        else:
            return None


ArgsKwargsInput = Union[ArgsKwargs, Dict, List, Any, None]
"""ArgsKwargsInput is the input type for ArgsKwargs"""

ArgsKwargsOutput = Union[ArgsKwargs, List, Any, None]
"""ArgsKwargsOutput is the output type for ArgsKwargs"""



import math
from typing import Optional

import torch
from torch import Tensor, nn

from analogvnn.nn.activation.Activation import Activation

__all__ = ['PReLU', 'ReLU', 'LeakyReLU']


class PReLU(Activation):
    """Implements the parametric rectified linear unit (PReLU) activation function.

    Attributes:
        alpha (float): the slope of the negative part of the activation function.
        _zero (Tensor): placeholder tensor of zero.
    """
    __constants__ = ['alpha', '_zero']
    alpha: nn.Parameter
    _zero: nn.Parameter

    def __init__(self, alpha: float):
        """initialize the parametric rectified linear unit (PReLU) activation function.

        Args:
            alpha (float): the slope of the negative part of the activation function.
        """
        super(PReLU, self).__init__()
        self.alpha = nn.Parameter(torch.tensor(alpha), requires_grad=False)
        self._zero = nn.Parameter(torch.tensor(0), requires_grad=False)

    def forward(self, x: Tensor) -> Tensor:
        """forward pass of the parametric rectified linear unit (PReLU) activation function.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return torch.minimum(self._zero, x) * self.alpha + torch.maximum(self._zero, x)

    def backward(self, grad_output: Optional[Tensor]) -> Optional[Tensor]:
        """backward pass of the parametric rectified linear unit (PReLU) activation function.

        Args:
            grad_output (Optional[Tensor]): the gradient of the output tensor.

        Returns:
            Optional[Tensor]: the gradient of the input tensor.
        """
        x = self.inputs
        grad = (x < 0).type(torch.float) * self.alpha + (x >= 0).type(torch.float)
        return grad_output * grad

    @staticmethod
    def initialise(tensor: Tensor) -> Tensor:
        """initialize the tensor using kaiming uniform initialization with gain associated
        with the parametric rectified linear unit (PReLU) activation function.

        Args:
            tensor (Tensor): the tensor to be initialized.

        Returns:
            Tensor: the initialized tensor.
        """
        return nn.init.kaiming_uniform(tensor, a=math.sqrt(5), nonlinearity="leaky_relu")

    @staticmethod
    def initialise_(tensor: Tensor) -> Tensor:
        """in-place initialize the tensor using kaiming uniform initialization with gain associated
        with the parametric rectified linear unit (PReLU) activation function.

        Args:
            tensor (Tensor): the tensor to be initialized.

        Returns:
            Tensor: the initialized tensor.
        """
        return nn.init.kaiming_uniform_(tensor, a=math.sqrt(5), nonlinearity="leaky_relu")


class ReLU(PReLU):
    """Implements the rectified linear unit (ReLU) activation function.
    Attributes:
        alpha (float): 0
    """

    def __init__(self):
        """initialize the rectified linear unit (ReLU) activation function.
        """
        super(ReLU, self).__init__(alpha=0)

    @staticmethod
    def initialise(tensor: Tensor) -> Tensor:
        """initialize the tensor using kaiming uniform initialization with gain associated
        with the rectified linear unit (ReLU) activation function.

        Args:
            tensor (Tensor): the tensor to be initialized.

        Returns:
            Tensor: the initialized tensor.
        """
        return nn.init.kaiming_uniform(tensor, a=math.sqrt(5), nonlinearity="relu")

    @staticmethod
    def initialise_(tensor: Tensor) -> Tensor:
        return nn.init.kaiming_uniform_(tensor, a=math.sqrt(5), nonlinearity="relu")


class LeakyReLU(PReLU):
    """Implements the leaky rectified linear unit (LeakyReLU) activation function.

    Attributes:
        alpha (float): 0.01
    """

    def __init__(self):
        """initialize the leaky rectified linear unit (LeakyReLU) activation function.
        """
        super(LeakyReLU, self).__init__(alpha=0.01)



from typing import Optional, Tuple

from torch import nn
from torch.utils.data import DataLoader

__all__ = ['train']


def train(
        model: nn.Module,
        train_loader: DataLoader,
        epoch: Optional[int] = None,
        test_run: bool = False
) -> Tuple[float, float]:
    """Train the model on the train set.

    Args:
        model (torch.nn.Module): the model to train.
        train_loader (DataLoader): the train set.
        epoch (int): the current epoch.
        test_run (bool): is it a test run.

    Returns:
        tuple: the loss and accuracy of the model on the train set.
    """
    model.train()
    total_loss = 0.0
    total_accuracy = 0
    total_size = 0
    if isinstance(train_loader, DataLoader):
        # noinspection PyTypeChecker
        dataset_size = len(train_loader.dataset)
    else:
        dataset_size = len(train_loader)

    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(model.device), target.to(model.device)

        # zero the parameter gradients
        model.zero_grad()
        model.optimizer.zero_grad()

        # forward + backward + optimize
        output = model(data)
        loss, accuracy = model.loss(output, target)

        model.backward()
        model.optimizer.step()

        # print statistics
        total_loss += loss.item() * len(data)
        total_accuracy += accuracy * len(data)
        total_size += len(data)

        print_mod = int(dataset_size / (len(data) * 5))
        if print_mod > 0 and batch_idx % print_mod == 0 and batch_idx > 0:
            print(
                f'Train Epoch:'
                f' {((epoch + 1) if epoch is not None else "")}'
                f' [{batch_idx * len(data)}/{dataset_size} ({100. * batch_idx / len(train_loader):.0f}%)]'
                f'\tLoss: {total_loss / total_size:.6f}'
                f'\tAccuracy: {total_accuracy / total_size * 100:.2f}%'
            )

        if test_run:
            break

    total_loss /= total_size
    total_accuracy /= total_size
    return total_loss, total_accuracy



from typing import Optional

import torch
from torch import Tensor, nn

from analogvnn.nn.activation.Activation import Activation

__all__ = ['SELU', 'ELU']


class SELU(Activation):
    """Implements the scaled exponential linear unit (SELU) activation function.

    Attributes:
        alpha (nn.Parameter): the alpha parameter.
        scale_factor (nn.Parameter): the scale factor parameter.
    """
    __constants__ = ['alpha', 'scale_factor']
    alpha: nn.Parameter
    scale_factor: nn.Parameter

    def __init__(self, alpha: float = 1.0507, scale_factor: float = 1.):
        """initialize the scaled exponential linear unit (SELU) activation function.

        Args:
            alpha (float): the alpha parameter.
            scale_factor (float): the scale factor parameter.
        """
        super(SELU, self).__init__()
        self.alpha = nn.Parameter(torch.tensor(alpha), requires_grad=False)
        self.scale_factor = nn.Parameter(torch.tensor(scale_factor), requires_grad=False)

    def forward(self, x: Tensor) -> Tensor:
        """forward pass of the scaled exponential linear unit (SELU) activation function.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return self.scale_factor * (
                (x <= 0).type(torch.float) * self.alpha * (torch.exp(x) - 1) +
                (x > 0).type(torch.float) * x
        )

    def backward(self, grad_output: Optional[Tensor]) -> Optional[Tensor]:
        """backward pass of the scaled exponential linear unit (SELU) activation function.

        Args:
            grad_output (Optional[Tensor]): the gradient of the output tensor.

        Returns:
            Optional[Tensor]: the gradient of the input tensor.
        """
        x = self.inputs
        grad = self.scale_factor * ((x < 0).type(torch.float) * self.alpha * torch.exp(x) + (x >= 0).type(torch.float))
        return grad_output * grad

    @staticmethod
    def initialise(tensor: Tensor) -> Tensor:
        """initialize the tensor using xavier uniform initialization with gain associated
        with the scaled exponential linear unit (SELU) activation function.

        Args:
            tensor (Tensor): the tensor to be initialized.

        Returns:
            Tensor: the initialized tensor.
        """
        return nn.init.xavier_uniform(tensor, gain=nn.init.calculate_gain('selu'))

    @staticmethod
    def initialise_(tensor: Tensor) -> Tensor:
        """in-place initialize the tensor using xavier uniform initialization with gain associated
        with the scaled exponential linear unit (SELU) activation function.

        Args:
            tensor (Tensor): the tensor to be initialized.

        Returns:
            Tensor: the initialized tensor.
        """
        return nn.init.xavier_uniform_(tensor, gain=nn.init.calculate_gain('selu'))


class ELU(SELU):
    """Implements the exponential linear unit (ELU) activation function.

    Attributes:
        alpha (nn.Parameter): 1.0507
        scale_factor (nn.Parameter): 1.
    """

    def __init__(self, alpha: float = 1.0507):
        """initialize the exponential linear unit (ELU) activation function.

        Args:
            alpha (float): the alpha parameter.
        """
        super(ELU, self).__init__(alpha=alpha, scale_factor=1.)



from __future__ import annotations

import functools
from typing import Union, Type, Callable, Sequence, Optional, TYPE_CHECKING

from torch import nn, Tensor

from analogvnn.backward.BackwardFunction import BackwardFunction
from analogvnn.backward.BackwardModule import BackwardModule
from analogvnn.graph.ArgsKwargs import ArgsKwargs, ArgsKwargsOutput
from analogvnn.utils.common_types import TENSORS

if TYPE_CHECKING:
    from analogvnn.graph.ModelGraph import ModelGraph

__all__ = ['Layer']


# https://github.com/pytorch/pytorch/pull/91819
def __nn_Module_init_updated__(function: Callable) -> Callable:
    """ Wrapper for nn.Module.__init__ to support multiple parent classes at same time.
    Args:
        function (Callable): nn.Module.__init__ function

    Returns:
        Callable: Wrapped function
    """

    # noinspection PyUnusedLocal
    def _temp(*args, **kwargs) -> ...:
        pass

    @functools.wraps(function)
    def new_function(self, *args, **kwargs):
        super_init = None
        next_mro_index = self.__class__.__mro__.index(nn.Module) + 1
        next_mro_class = self.__class__.__mro__[next_mro_index]

        if next_mro_class is not object:
            super_init = next_mro_class.__init__
            next_mro_class.__init__ = _temp

        function(self, *args, **kwargs)

        if next_mro_class is not object:
            next_mro_class.__init__ = super_init
            super(nn.Module, self).__init__()

    return new_function


nn.Module.__init__ = __nn_Module_init_updated__(nn.Module.__init__)
""" nn.Module.__init__ is updated to support multiple parent classes at same time. """


class Layer(nn.Module):
    """Base class for analog neural network modules.

    Attributes:
        _inputs (Union[None, ArgsKwargs]): Inputs of the layer.
        _outputs (Union[None, Tensor, Sequence[Tensor]]): Outputs of the layer.
        _backward_module (Optional[BackwardModule]): Backward module of the layer.
        _use_autograd_graph (bool): If True, the autograd graph is used to calculate the gradients.
        graphs (Optional[ModelGraph]): Contains Forward and Backward Graphs of the layer.
    """
    _inputs: Union[None, ArgsKwargs]
    _outputs: Union[None, Tensor, Sequence[Tensor]]
    _backward_module: Optional[BackwardModule]
    _use_autograd_graph: bool
    graphs: Optional[ModelGraph]

    def __init__(self):
        """Initializes the layer."""
        super(Layer, self).__init__()
        self._inputs = None
        self._outputs = None
        self._backward_module = None
        self._use_autograd_graph = False
        self.graphs = None

    def __call__(self, *inputs, **kwargs):
        """Calls the forward pass of neural network layer.

        Args:
            *inputs: Inputs of the forward pass.
            **kwargs: Keyword arguments of the forward pass.
        """
        self._forward_wrapper(self.forward)
        outputs = super(Layer, self).__call__(*inputs, **kwargs)
        if self.training:
            self._inputs = ArgsKwargs(args=inputs, kwargs=kwargs)
            self._outputs = outputs

        return outputs

    @property
    def use_autograd_graph(self) -> bool:
        """If True, the autograd graph is used to calculate the gradients.

        Returns:
            bool: use_autograd_graph.
        """
        if self.graphs is not None:
            return self.graphs.use_autograd_graph
        return self._use_autograd_graph

    @use_autograd_graph.setter
    def use_autograd_graph(self, use_autograd_graph: bool):
        """Sets the use_autograd_graph attribute.

        Args:
            use_autograd_graph (bool): use_autograd_graph.
        """
        self._use_autograd_graph = use_autograd_graph
        if self.graphs is not None:
            self.graphs.use_autograd_graph = use_autograd_graph

    @property
    def inputs(self) -> ArgsKwargsOutput:
        """Inputs of the layer.

        Returns:
            ArgsKwargsOutput: inputs.
        """
        return ArgsKwargs.from_args_kwargs_object(self._inputs)

    @property
    def outputs(self) -> Union[None, Tensor, Sequence[Tensor]]:
        """Outputs of the layer.

        Returns:
            Union[None, Tensor, Sequence[Tensor]]: outputs.
        """
        return self._outputs

    @property
    def backward_function(self) -> Union[None, Callable, BackwardModule]:
        """Backward module of the layer.

        Returns:
            Union[None, Callable, BackwardModule]: backward_function.
        """
        if self._backward_module is not None:
            return self._backward_module

        if isinstance(self, BackwardModule):
            return self

        return None

    @backward_function.setter
    def backward_function(self, function: Union[BackwardModule, Type[BackwardModule], Callable]):
        """Sets the backward_function attribute.

        Args:
            function (Union[BackwardModule, Type[BackwardModule], Callable]): backward_function.
        """
        self.set_backward_function(function)

    def set_backward_function(self, backward_class: Union[Callable, BackwardModule, Type[BackwardModule]]) -> Layer:
        """Sets the backward_function attribute.

        Args:
            backward_class (Union[Callable, BackwardModule, Type[BackwardModule]]): backward_function.

        Returns:
            Layer: self.

        Raises:
            TypeError: If backward_class is not a callable or BackwardModule.
        """
        if backward_class == self:
            return self

        if issubclass(backward_class, BackwardModule):
            self._backward_module = backward_class(self)
        elif isinstance(backward_class, BackwardModule):
            backward_class.set_layer(self)
            self._backward_module = backward_class
        elif callable(backward_class):
            self._backward_module = BackwardFunction(backward_class, self)
        else:
            raise TypeError(f"Backward Module is not set for '{self}'")

        return self

    def _forward_wrapper(self, function: Callable) -> Callable:
        """Wrapper for the forward function.

        Args:
            function (Callable): Forward function.

        Returns:
            Callable: Wrapped function.
        """
        # noinspection PyUnresolvedReferences
        if hasattr(function, "__wrapper__") and function.__wrapper__ == Layer._forward_wrapper:
            return function

        if not isinstance(self.backward_function, BackwardModule):
            return function

        if not self.backward_function.has_forward():
            self.backward_function.forward = self.forward

        @functools.wraps(function)
        def new_forward(*args, **kwargs):
            return self.backward_function.auto_apply(*args, to_apply=self.use_autograd_graph, **kwargs)

        new_forward.__wrapped__ = function
        new_forward.__wrapper__ = Layer._forward_wrapper
        setattr(self, "forward", new_forward)
        return new_forward

    def _call_impl_forward(self, *args: Tensor, **kwargs: Tensor) -> TENSORS:
        """Calls the forward pass of the layer.

        Args:
            *args: Inputs of the forward pass.
            **kwargs: Keyword arguments of the forward pass.

        Returns:
            TENSORS: Outputs of the forward pass.
        """
        if isinstance(self.backward_function, BackwardModule) and self.backward_function.has_forward():
            forward_functions = self.backward_function.forward
        else:
            forward_functions = self.forward

        if hasattr(forward_functions, "__wrapped__"):
            forward_functions = forward_functions.__wrapped__

        return forward_functions(*args, **kwargs)



from typing import Optional, Tuple

import torch
from torch import Tensor, nn

from analogvnn.backward.BackwardIdentity import BackwardIdentity
from analogvnn.nn.module.Layer import Layer
from analogvnn.utils.common_types import TENSOR_OPERABLE
from analogvnn.utils.to_tensor_parameter import to_float_tensor, to_nongrad_parameter

__all__ = ['UniformNoise']


class UniformNoise(Layer, BackwardIdentity):
    """Implements the uniform noise function.

    Attributes:
        low (nn.Parameter): the lower bound of the uniform noise.
        high (nn.Parameter): the upper bound of the uniform noise.
        leakage (nn.Parameter): the leakage of the uniform noise.
        precision (nn.Parameter): the precision of the uniform noise.
    """
    __constants__ = ['low', 'high', 'leakage', 'precision']
    low: nn.Parameter
    high: nn.Parameter
    leakage: nn.Parameter
    precision: nn.Parameter

    def __init__(
            self,
            low: Optional[float] = None,
            high: Optional[float] = None,
            leakage: Optional[float] = None,
            precision: Optional[int] = None
    ):
        """initialize the uniform noise function.

        Args:
            low (float): the lower bound of the uniform noise.
            high (float): the upper bound of the uniform noise.
            leakage (float): the leakage of the uniform noise.
            precision (int): the precision of the uniform noise.
        """
        super(UniformNoise, self).__init__()

        if (low is None or high is None) + (leakage is None) + (precision is None) != 1:
            raise ValueError("only 2 out of 3 arguments are needed (scale, leakage, precision)")

        low, high, leakage, precision = to_float_tensor(low, high, leakage, precision)

        if (low is None or high is None) and leakage is not None and precision is not None:
            low, high = self.calc_high_low(leakage, precision)

        if low is not None and high is not None and leakage is None and precision is not None:
            leakage = self.calc_leakage(low, high, precision)

        if low is not None and high is not None and leakage is not None and precision is None:
            precision = self.calc_precision(low, high, leakage)

        self.low, self.high, self.leakage, self.precision = to_nongrad_parameter(low, high, leakage, precision)

    @staticmethod
    def calc_high_low(leakage: TENSOR_OPERABLE, precision: TENSOR_OPERABLE) -> Tuple[TENSOR_OPERABLE, TENSOR_OPERABLE]:
        """calculate the high and low from leakage and precision.

        Args:
            leakage (TENSOR_OPERABLE): the leakage of the uniform noise.
            precision (TENSOR_OPERABLE): the precision of the uniform noise.

        Returns:
            Tuple[TENSOR_OPERABLE, TENSOR_OPERABLE]: the high and low of the uniform noise.
        """
        v = 1 / (1 - leakage) * (1 / precision)
        return -v / 2, v / 2

    @staticmethod
    def calc_leakage(low: TENSOR_OPERABLE, high: TENSOR_OPERABLE, precision: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """calculate the leakage from low, high and precision.

        Args:
            low (TENSOR_OPERABLE): the lower bound of the uniform noise.
            high (TENSOR_OPERABLE): the upper bound of the uniform noise.
            precision (TENSOR_OPERABLE): the precision of the uniform noise.

        Returns:
            TENSOR_OPERABLE: the leakage of the uniform noise.
        """
        return 1 - min(1, (1 / precision) * (1 / (high - low)))

    @staticmethod
    def calc_precision(low: TENSOR_OPERABLE, high: TENSOR_OPERABLE, leakage: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """calculate the precision from low, high and leakage.

        Args:
            low (TENSOR_OPERABLE): the lower bound of the uniform noise.
            high (TENSOR_OPERABLE): the upper bound of the uniform noise.
            leakage (TENSOR_OPERABLE): the leakage of the uniform noise.

        Returns:
            TENSOR_OPERABLE: the precision of the uniform noise.
        """
        return 1 / (1 - leakage) * (1 / (high - low))

    @property
    def mean(self) -> Tensor:
        """the mean of the uniform noise.

        Returns:
            Tensor: the mean of the uniform noise.
        """
        return (self.high + self.low) / 2

    @property
    def stddev(self) -> Tensor:
        """the standard deviation of the uniform noise.

        Returns:
            Tensor: the standard deviation of the uniform noise.
        """
        return (self.high - self.low) / 12 ** 0.5

    @property
    def variance(self) -> Tensor:
        """the variance of the uniform noise.

        Returns:
            Tensor: the variance of the uniform noise.
        """
        return (self.high - self.low).pow(2) / 12

    def pdf(self, x: Tensor) -> Tensor:
        """the probability density function of the uniform noise.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the probability density function of the uniform noise.
        """
        return torch.exp(self.log_prob(x=x))

    def log_prob(self, x: Tensor) -> Tensor:
        """the log probability density function of the uniform noise.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the log probability density function of the uniform noise.
        """
        lb = self.low.le(x).type_as(self.low)
        ub = self.high.gt(x).type_as(self.low)
        return torch.log(lb.mul(ub)) - torch.log(self.high - self.low)

    def cdf(self, x: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """the cumulative distribution function of the uniform noise.

        Args:
            x (TENSOR_OPERABLE): the input tensor.

        Returns:
            TENSOR_OPERABLE: the cumulative distribution function of the uniform noise.
        """
        result = (x - self.low) / (self.high - self.low)
        return result.clamp(min=0, max=1)

    def forward(self, x: Tensor) -> Tensor:
        """add the uniform noise to the input tensor.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return torch.distributions.Uniform(low=x + self.low, high=x + self.high).sample()

    def extra_repr(self) -> str:
        """the extra representation of the uniform noise.

        Returns:
            str: the extra representation of the uniform noise.
        """
        return f'high={float(self.high):.4f}' \
               f', low={float(self.low):.4f}' \
               f', leakage={float(self.leakage):.4f}' \
               f', precision={int(self.precision)}'



import math
from numbers import Real
from typing import Optional

import scipy.special
import torch
from torch import Tensor, nn

from analogvnn.backward.BackwardIdentity import BackwardIdentity
from analogvnn.nn.module.Layer import Layer
from analogvnn.utils.common_types import TENSOR_OPERABLE
from analogvnn.utils.to_tensor_parameter import to_nongrad_parameter, to_float_tensor

__all__ = ['GaussianNoise']


class GaussianNoise(Layer, BackwardIdentity):
    """Implements the Gaussian noise function.
    
    Attributes:
        std (nn.Parameter): the standard deviation of the Gaussian noise.
        leakage (nn.Parameter): the leakage of the Gaussian noise.
        precision (nn.Parameter): the precision of the Gaussian noise.
    """
    __constants__ = ['std', 'leakage', 'precision']
    std: nn.Parameter
    leakage: nn.Parameter
    precision: nn.Parameter

    def __init__(
            self,
            std: Optional[float] = None,
            leakage: Optional[float] = None,
            precision: Optional[int] = None
    ):
        """initialize the Gaussian noise function.
        
        Args:
            std (float): the standard deviation of the Gaussian noise.
            leakage (float): the leakage of the Gaussian noise.
            precision (int): the precision of the Gaussian noise.
        """
        super(GaussianNoise, self).__init__()

        if (std is None) + (leakage is None) + (precision is None) != 1:
            raise ValueError("only 2 out of 3 arguments are needed (std, leakage, precision)")

        std, leakage, precision = to_float_tensor(std, leakage, precision)

        if std is None and leakage is not None and precision is not None:
            std = self.calc_std(leakage, precision)

        if precision is None and std is not None and leakage is not None:
            precision = self.calc_precision(std, leakage)

        if leakage is None and std is not None and precision is not None:
            leakage = self.calc_leakage(std, precision)

        self.std, self.leakage, self.precision = to_nongrad_parameter(std, leakage, precision)

    @staticmethod
    def calc_std(leakage: TENSOR_OPERABLE, precision: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """calculate the standard deviation of the Gaussian noise.
        
        Args:
            leakage (float): the leakage of the Gaussian noise.
            precision (int): the precision of the Gaussian noise.
            
        Returns:
            float: the standard deviation of the Gaussian noise.
        """
        return 1 / (2 * precision * scipy.special.erfinv(1 - leakage) * math.sqrt(2))

    @staticmethod
    def calc_precision(std: TENSOR_OPERABLE, leakage: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """calculate the precision of the Gaussian noise.
        
        Args:
            std (float): the standard deviation of the Gaussian noise.
            leakage (float): the leakage of the Gaussian noise.
            
        Returns:
            int: the precision of the Gaussian noise.
        """
        return 1 / (2 * std * scipy.special.erfinv(1 - leakage) * math.sqrt(2))

    @staticmethod
    def calc_leakage(std: TENSOR_OPERABLE, precision: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """calculate the leakage of the Gaussian noise.
        
        Args:
            std (float): the standard deviation of the Gaussian noise.
            precision (int): the precision of the Gaussian noise.
            
        Returns:
            float: the leakage of the Gaussian noise.
        """
        return 2 * GaussianNoise.static_cdf(x=-1 / (2 * precision), std=std)

    @property
    def stddev(self) -> Tensor:
        """the standard deviation of the Gaussian noise.
        
        Returns:
            Tensor: the standard deviation of the Gaussian noise.
        """
        return self.std

    @property
    def variance(self) -> Tensor:
        """the variance of the Gaussian noise.
        
        Returns:
            Tensor: the variance of the Gaussian noise.
        """
        return self.stddev.pow(2)

    def pdf(self, x: Tensor, mean: Tensor = 0) -> Tensor:
        """calculate the probability density function of the Gaussian noise.
        
        Args:
            x (Tensor): the input tensor.
            mean (Tensor): the mean of the Gaussian noise.
            
        Returns:
            Tensor: the probability density function of the Gaussian noise.
        """
        return torch.exp(self.log_prob(x=x, mean=mean))

    def log_prob(self, x: Tensor, mean: Tensor = 0) -> Tensor:
        """calculate the log probability density function of the Gaussian noise.
        
        Args:
            x (Tensor): the input tensor.
            mean (Tensor): the mean of the Gaussian noise.
            
        Returns:
            Tensor: the log probability density function of the Gaussian noise.
        """
        x = x if isinstance(x, Tensor) else torch.tensor(x, requires_grad=False)
        mean = mean if isinstance(mean, Tensor) else torch.tensor(mean, requires_grad=False)

        var = (self.std ** 2)
        log_scale = math.log(self.std) if isinstance(self.std, Real) else self.std.log()
        return -((x - mean) ** 2) / (2 * var) - log_scale - math.log(math.sqrt(2 * math.pi))

    @staticmethod
    def static_cdf(x: TENSOR_OPERABLE, std: TENSOR_OPERABLE, mean: TENSOR_OPERABLE = 0.) -> TENSOR_OPERABLE:
        """calculate the cumulative distribution function of the Gaussian noise.
        
        Args:
            x (TENSOR_OPERABLE): the input tensor.
            std (TENSOR_OPERABLE): the standard deviation of the Gaussian noise.
            mean (TENSOR_OPERABLE): the mean of the Gaussian noise.
            
        Returns:
            TENSOR_OPERABLE: the cumulative distribution function of the Gaussian noise.
        """
        return 1 / 2 * (1 + math.erf((x - mean) / (std * math.sqrt(2))))

    def cdf(self, x: Tensor, mean: Tensor = 0) -> Tensor:
        """calculate the cumulative distribution function of the Gaussian noise.
        
        Args:
            x (Tensor): the input tensor.
            mean (Tensor): the mean of the Gaussian noise.
            
        Returns:
            Tensor: the cumulative distribution function of the Gaussian noise.
        """
        x = x if isinstance(x, Tensor) else torch.tensor(x, requires_grad=False)
        mean = mean if isinstance(mean, Tensor) else torch.tensor(mean, requires_grad=False)
        return self.static_cdf(x=x, std=self.std, mean=mean)

    def forward(self, x: Tensor) -> Tensor:
        """add the Gaussian noise to the input tensor.
        
        Args:
            x (Tensor): the input tensor.
            
        Returns:
            Tensor: the output tensor.
        """
        return torch.normal(mean=x, std=self.std)

    def extra_repr(self) -> str:
        """the extra representation of the Gaussian noise.
        
        Returns:
            str: the extra representation of the Gaussian noise.
        """
        return f'std={float(self.std):.4f}, leakage={float(self.leakage):.4f}, precision={int(self.precision)}'



import torch
from torch import Tensor

from analogvnn.utils.common_types import TENSOR_OPERABLE

__all__ = ['reduce_precision', 'stochastic_reduce_precision']


def reduce_precision(x: TENSOR_OPERABLE, precision: TENSOR_OPERABLE, divide: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
    """`reduce_precision` takes `x` and reduces its precision to `precision` by rounding to the
    nearest multiple of `precision`.
    
    Args:
      x (TENSOR_OPERABLE): Tensor
      precision (TENSOR_OPERABLE): the precision of the quantization.
      divide (TENSOR_OPERABLE): the number of bits to be reduced
    
    Returns:
      TENSOR_OPERABLE: TENSOR_OPERABLE with the same shape as x, but with values rounded to the nearest
      multiple of precision.
    """
    x = x if isinstance(x, Tensor) else torch.tensor(x, requires_grad=False)
    g: Tensor = x * precision
    f = torch.sign(g) * torch.maximum(
        torch.floor(torch.abs(g)),
        torch.ceil(torch.abs(g) - divide)
    ) * (1 / precision)
    return f


def stochastic_reduce_precision(x: TENSOR_OPERABLE, precision: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
    """`stochastic_reduce_precision` takes `x` and reduces its precision to `precision` by rounding to the
    nearest multiple of `precision` with a stochastic rounding scheme.
    
    Args:
        x (TENSOR_OPERABLE): Tensor
        precision (TENSOR_OPERABLE): the precision of the quantization.

    Returns:
        TENSOR_OPERABLE: TENSOR_OPERABLE with the same shape as x, but with values rounded to the
        nearest multiple of precision.
    """
    g: Tensor = x * precision
    rand_x = torch.rand_like(g, requires_grad=False)

    g_abs = torch.abs(g)
    g_floor = torch.floor(g_abs)
    g_ceil = torch.ceil(g_abs)

    prob_floor = 1 - torch.abs(g_floor - g_abs)
    bool_floor = rand_x <= prob_floor
    do_floor = bool_floor.type(torch.float)
    do_ceil = torch.logical_not(bool_floor).type(torch.float)

    f = torch.sign(g) * (do_floor * g_floor + do_ceil * g_ceil) * (1 / precision)
    return f



from torch import nn

__all__ = ['Parameter']


class Parameter(nn.Parameter):
    """A parameter that can be used in a `torch.nn.Module`.

    This class is a wrapper around `torch.nn.Parameter` that allows to set additional attributes.
    """

    def __new__(cls, data=None, requires_grad=True, *args, **kwargs):
        """Creates a new parameter.

        Args:
            data: the data for the parameter.
            requires_grad (bool): whether the parameter requires gradient.
            *args: additional arguments.
            **kwargs: additional keyword arguments.

        Returns:
            Parameter: the created parameter.
        """
        return super(Parameter, cls).__new__(cls, data, requires_grad)

    # noinspection PyUnusedLocal
    def __init__(self, data=None, requires_grad=True, *args, **kwargs):
        """Initializes the parameter.

        Args:
            data: the data for the parameter.
            requires_grad (bool): whether the parameter requires gradient.
            *args: additional arguments.
            **kwargs: additional keyword arguments.
        """
        super(Parameter, self).__init__()

    def __repr__(self, *args, **kwargs):
        """Returns a string representation of the parameter.

        Args:
            *args: additional arguments.
            **kwargs: additional keyword arguments.

        Returns:
            str: the string representation.
        """
        return super(Parameter, self).__repr__(*args, **kwargs)



from typing import Tuple, Union

import torch
from torch import nn

__all__ = ['to_float_tensor', 'to_nongrad_parameter']


def to_float_tensor(*args) -> Tuple[Union[torch.Tensor, None], ...]:
    """Converts the given arguments to `torch.Tensor` of type `torch.float32`.

    The returned tensors are not trainable.

    Args:
        *args: the arguments to convert.

    Returns:
        tuple: the converted arguments.
    """
    return tuple((None if i is None else torch.tensor(i, requires_grad=False, dtype=torch.float)) for i in args)


def to_nongrad_parameter(*args) -> Tuple[Union[nn.Parameter, None], ...]:
    """Converts the given arguments to `nn.Parameter` of type `torch.float32`.

    The returned parameters are not trainable.

    Args:
        *args: the arguments to convert.

    Returns:
        tuple: the converted arguments.
    """
    return tuple((None if i is None else nn.Parameter(i, requires_grad=False)) for i in to_float_tensor(*args))



from analogvnn.backward.BackwardIdentity import BackwardIdentity
from analogvnn.nn.module.Layer import Layer

__all__ = ['Normalize']


class Normalize(Layer, BackwardIdentity):
    """This class is base class for all normalization functions.
    """
    pass



from __future__ import annotations

import dataclasses
import typing
import warnings
from collections import namedtuple
from dataclasses import dataclass
from distutils.version import LooseVersion
from typing import Optional, Sequence, List, Dict, Union, Any, Callable

import torch
from torch import Tensor, nn

from analogvnn.nn.module.Layer import Layer
from analogvnn.utils.is_cpu_cuda import is_cpu_cuda

if typing.TYPE_CHECKING:
    from graphviz import Digraph

__all__ = [
    'save_autograd_graph',
    'save_autograd_graph_from_trace',
    'get_autograd_dot',
    'get_autograd_dot_from_trace',
    'get_autograd_obj',
    'make_autograd_obj',
    'size_to_str',
    'AutoGradDot',
]

Node = namedtuple('Node', ('name', 'inputs', 'attr', 'op'))

# Saved attrs for grad_fn (incl. saved variables) begin with `._saved_*`
SAVED_PREFIX = "_saved_"


def size_to_str(size):
    """Convert a tensor size to a string.

    Args:
        size (torch.Size): the size to convert.

    Returns:
        str: the string representation of the size.
    """
    return '(' + ', '.join(['%d' % s for s in size]) + ')'


def resize_graph(dot: Digraph, size_per_element: float = 0.15, min_size: float = 12):
    """Resize the graph according to how much content it contains.

    Modify the graph in place.

    Args:
        dot (graphviz.Digraph): graph to be resized
        size_per_element (float): A "rank" in graphviz contains roughly
            size_per_element**2 pixels.
        min_size (float): Minimum size of graph.
    """
    # Get the approximate number of nodes and edges
    num_rows = len(dot.body)
    content_size = num_rows * size_per_element
    size = max(min_size, content_size)
    size_str = str(size) + "," + str(size)
    dot.graph_attr.update(size=size_str)


def get_fn_name(fn: Callable, show_attrs: bool, max_attr_chars: int) -> str:
    """Get the name of a function.

    Args:
        fn (Callable): the function.
        show_attrs (bool): whether to show the attributes.
        max_attr_chars (int): the maximum number of characters to show for the attributes.

    Returns:
        str: the name of the function.
    """

    name = str(type(fn).__name__)
    if not show_attrs:
        return name
    attrs = dict()
    for attr in dir(fn):
        if not attr.startswith(SAVED_PREFIX):
            continue
        val = getattr(fn, attr)
        attr = attr[len(SAVED_PREFIX):]
        if torch.is_tensor(val):
            attrs[attr] = "[saved tensor]"
        elif isinstance(val, tuple) and any(torch.is_tensor(t) for t in val):
            attrs[attr] = "[saved tensors]"
        else:
            attrs[attr] = str(val)
    if not attrs:
        return name
    max_attr_chars = max(max_attr_chars, 3)
    col1width = max(len(k) for k in attrs.keys())
    col2width = min(max(len(str(v)) for v in attrs.values()), max_attr_chars)
    sep = "-" * max(col1width + col2width + 2, len(name))
    attrstr = '%-' + str(col1width) + 's: %' + str(col2width) + 's'
    truncate = lambda s: s[:col2width - 3] + "..." if len(s) > col2width else s
    params = '\n'.join(attrstr % (k, truncate(str(v))) for (k, v) in attrs.items())
    return name + '\n' + sep + '\n' + params


@dataclass
class AutoGradDot:
    """Stores and manages Graphviz representation of PyTorch autograd graph.

    Attributes:
        dot (graphviz.Digraph): Graphviz representation of the autograd graph.
        _module (nn.Module): The module to be traced.
        _inputs (List[Tensor]): The inputs to the module.
        _inputs_kwargs (Dict[str, Tensor]): The keyword arguments to the module.
        _outputs (Sequence[Tensor]): The outputs of the module.
        param_map (Dict[int, str]): A map from parameter values to their names.
        _seen (set): A set of nodes that have already been added to the graph.
        show_attrs (bool): whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved (bool): whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars (int): if show_attrs is `True`, sets max number of characters
            to display for any given attribute.
        _called (bool): the module has been called.
    """

    dot: Digraph = None
    _module: nn.Module = None

    _inputs: Sequence[Tensor] = dataclasses.field(default=None, repr=False, hash=False)
    _inputs_kwargs: Dict[str, Tensor] = dataclasses.field(default_factory=dict, repr=False, hash=False)
    _outputs: Sequence[Tensor] = dataclasses.field(default=None, repr=False, hash=False)

    param_map: dict = dataclasses.field(default_factory=dict, repr=False, hash=False)
    _seen: set = dataclasses.field(default_factory=set, repr=False, hash=False)

    show_attrs: bool = dataclasses.field(default=False, repr=False, hash=False)
    show_saved: bool = dataclasses.field(default=False, repr=False, hash=False)
    max_attr_chars: int = dataclasses.field(default=50, repr=False, hash=False)
    _called: bool = False

    def __post_init__(self):
        """Create the graphviz graph.

        Raises:
            ImportError: if graphviz (https://pygraphviz.github.io/) is not available.
        """

        try:
            from graphviz import Digraph
        except ImportError as e:
            raise ImportError("requires graphviz: https://pygraphviz.github.io/") from e

        node_attr = dict(
            style='filled',
            shape='box',
            align='left',
            fontsize='12',
            ranksep='0.1',
            height='0.2',
            fontname='monospace'
        )
        self.dot = Digraph(node_attr=node_attr, graph_attr=dict(size="12,12"), format="svg")

    def __call__(self) -> Optional[Sequence[Tensor]]:
        """Call the module with inputs and returns the outputs.

        Returns:
            Optional[Sequence[Tensor]]: the outputs of the module.
        """

        assert self._module is not None
        outputs = self._module(*self.inputs, **self.inputs_kwargs)

        if outputs is None:
            return outputs

        if isinstance(outputs, tuple):
            return outputs

        return (outputs,)

    @property
    def inputs(self) -> Sequence[Tensor]:
        """the arg inputs to the module

        Returns:
            Sequence[Tensor]: the arg inputs to the module.
        """
        return self._inputs

    @inputs.setter
    def inputs(self, inputs: Union[Tensor, Sequence[Tensor]]):
        """Set the inputs to the module.

        Args:
            inputs (Union[Tensor, Sequence[Tensor]]): the inputs to the module.
        """
        self._inputs = inputs
        self._called = False

        if not inputs:
            return

        if not isinstance(inputs, tuple):
            inputs = (inputs,)

        for i, v in enumerate(inputs):
            self.param_map[id(v)] = f"INPUT_{i}"
            self.param_map[id(v.data)] = f"INPUT_{i}"

    @property
    def inputs_kwargs(self) -> Dict[str, Tensor]:
        """the keyword inputs to the module

        Args:
            Dict[str, Tensor]: the keyword inputs to the module.
        """
        return self._inputs_kwargs

    @inputs_kwargs.setter
    def inputs_kwargs(self, inputs_kwargs: Dict[str, Tensor]):
        """Set the keyword inputs to the module.

        Args:
            inputs_kwargs (Dict[str, Tensor]): the keyword inputs to the module.
        """
        self._inputs_kwargs = inputs_kwargs
        self._called = False

        if not inputs_kwargs:
            return

        for k, v in inputs_kwargs.items():
            self.param_map[id(v)] = f"INPUT_{k}"
            self.param_map[id(v.data)] = f"INPUT_{k}"

    @property
    def outputs(self) -> Optional[Sequence[Tensor]]:
        """the outputs of the module

        Returns:
            Optional[Sequence[Tensor]]: the outputs of the module.
        """
        if self._called:
            return self._outputs

        outputs = self.__call__()
        self._called = True

        if not outputs:
            return

        for i, v in enumerate(outputs):
            self.param_map[id(v)] = f"OUTPUT_{i}"
            self.param_map[id(v.data)] = f"OUTPUT_{i}"

        self._outputs = outputs
        return self._outputs

    @property
    def module(self):
        """the module.

        Returns:
            nn.Module: the module to be traced.
        """
        return self._module

    @module.setter
    def module(self, module: nn.Module):
        """Set the module.

        Args:
            module (nn.Module): the module.
        """
        self._module = module
        self._called = False

        if not module:
            return

        assert isinstance(module, nn.Module)
        for k, v in dict(module.named_parameters()).items():
            self.param_map[id(v)] = k
            self.param_map[id(v.data)] = k

    def reset_params(self):
        """Reset the param_map and _seen.

        Returns:
            AutoGradDot: self.
        """
        self.param_map = {}
        self._seen = set()
        self.inputs = self.inputs
        self.inputs_kwargs = self.inputs_kwargs
        self.module = self.module
        return self

    def get_tensor_name(self, tensor: Tensor, name: Optional[str] = None):
        """Get the name of the tensor.

        Args:
            tensor (Tensor): the tensor.
            name (Optional[str]): the name of the tensor. Defaults to None.

        Returns:
            str: the name of the tensor.
        """
        if not name:
            if id(tensor) in self.param_map:
                name = self.param_map[id(tensor)]
            elif hasattr(tensor, "name") and tensor.name is not None:
                name = tensor.name
            elif hasattr(tensor, "names") and tensor.names is not None and len(tensor.names) > 0:
                name = str(tensor.names)
            else:
                name = ''
        return '%s\n %s' % (name, size_to_str(tensor.size()))

    def add_tensor(self, tensor: Tensor, name: Optional[str] = None, _attributes=None, **kwargs):
        """Add a tensor to the graph.

        Args:
            tensor (Tensor): the tensor.
            name (Optional[str]): the name of the tensor. Defaults to None.
            _attributes (Optional[Dict[str, str]]): the attributes of the tensor. Defaults to None.
            **kwargs: the attributes of the dot.node function.

        Returns:
            AutoGradDot: self.
        """
        self._seen.add(tensor)
        self.dot.node(
            name=str(id(tensor)),
            label=self.get_tensor_name(tensor, name=name),
            _attributes=_attributes,
            **kwargs
        )
        return self

    def add_fn(self, fn: Any, _attributes=None, **kwargs):
        """Add a function to the graph.

        Args:
            fn (Any): the function.
            _attributes (Optional[Dict[str, str]]): the attributes of the function. Defaults to None.
            **kwargs: the attributes of the dot.node function.

        Returns:
            AutoGradDot: self.
        """
        self._seen.add(fn)
        self.dot.node(
            name=str(id(fn)),
            label=get_fn_name(fn, self.show_attrs, self.max_attr_chars),
            _attributes=_attributes,
            **kwargs
        )
        return self

    def add_edge(self, u: Any, v: Any, label: Optional[str] = None, _attributes=None, **kwargs):
        """Add an edge to the graph.

        Args:
            u (Any): tail node.
            v (Any): head node.
            label (Optional[str]): the label of the edge. Defaults to None.
            _attributes (Optional[Dict[str, str]]): the attributes of the edge. Defaults to None.
            **kwargs: the attributes of the dot.edge function.

        Returns:
            AutoGradDot: self.
        """
        self.dot.edge(
            tail_name=str(id(u)),
            head_name=str(id(v)),
            label=label,
            _attributes=_attributes,
            **kwargs
        )
        return self

    def add_seen(self, item: Any):
        """Add an item to the seen set.

        Args:
            item (Any): the item.

        Returns:
            AutoGradDot: self.
        """
        self._seen.add(item)
        return self

    def is_seen(self, item: Any) -> bool:
        """Check if the item is in the seen set.

        Args:
            item (Any): the item.

        Returns:
            bool: True if the item is in the seen set.
        """
        return item in self._seen


def add_grad_fn(grad_fn, autograd_dot: AutoGradDot):
    """Add a grad_fn to the graph.

    Args:
        grad_fn (Any): the Tensor.grad_fn.
        autograd_dot (AutoGradDot): the AutoGradDot object.
    """
    assert not torch.is_tensor(grad_fn)
    if autograd_dot.is_seen(grad_fn):
        return

    # add the node for this grad_fn
    autograd_dot.add_fn(grad_fn)

    if autograd_dot.show_saved:
        for attr in dir(grad_fn):
            if not attr.startswith(SAVED_PREFIX):
                continue
            val = getattr(grad_fn, attr)
            autograd_dot.add_seen(val)
            attr = attr[len(SAVED_PREFIX):]
            if torch.is_tensor(val):
                autograd_dot.add_edge(grad_fn, val, dir="none")
                autograd_dot.add_tensor(val, name=attr, fillcolor='orange')
            if isinstance(val, tuple):
                for i, t in enumerate(val):
                    if torch.is_tensor(t):
                        name = attr + '[%s]' % str(i)
                        autograd_dot.add_edge(grad_fn, t, dir="none")
                        autograd_dot.add_tensor(t, name=name, fillcolor='orange')

    if hasattr(grad_fn, 'variable'):
        # if grad_accumulator, add the node for `.variable`
        var = grad_fn.variable
        autograd_dot.add_tensor(var, fillcolor='lightblue')
        autograd_dot.add_edge(var, grad_fn)

    # recurse
    if hasattr(grad_fn, 'next_functions'):
        for u in grad_fn.next_functions:
            if u[0] is not None:
                autograd_dot.add_edge(u[0], grad_fn)
                add_grad_fn(u[0], autograd_dot=autograd_dot)

    # note: this used to show .saved_tensors in pytorch0.2, but stopped
    # working* as it was moved to ATen and Variable-Tensor merged
    # also note that this still works for custom autograd functions
    if hasattr(grad_fn, 'saved_tensors'):
        for t in grad_fn.saved_tensors:
            autograd_dot.add_edge(t, grad_fn)
            autograd_dot.add_tensor(t, fillcolor='orange')


def add_base_tensor(tensor: Tensor, autograd_dot: AutoGradDot, color: str = 'darkolivegreen1'):
    """Add a base tensor to the graph.

    Args:
        tensor (Tensor): the base tensor.
        autograd_dot (AutoGradDot): the AutoGradDot object.
        color (str): the color of the base tensor in graph. Defaults to 'darkolivegreen1'.
    """
    if autograd_dot.is_seen(tensor):
        return

    autograd_dot.add_tensor(tensor, fillcolor=color)

    if tensor.grad_fn:
        add_grad_fn(tensor.grad_fn, autograd_dot=autograd_dot)
        autograd_dot.add_edge(tensor.grad_fn, tensor)

    if tensor._is_view():
        add_base_tensor(tensor._base, autograd_dot=autograd_dot, color='darkolivegreen3')
        autograd_dot.add_edge(tensor._base, tensor, style="dotted")


def make_autograd_obj(
        module: nn.Module,
        *args: Tensor,
        additional_params: Optional[dict] = None,
        show_attrs: bool = True,
        show_saved: bool = True,
        max_attr_chars: int = 50,
        **kwargs: Tensor
) -> AutoGradDot:
    """Produces Graphviz representation of PyTorch autograd graph.

    If a node represents a backward function, it is gray. Otherwise, the node
    represents a tensor and is either blue, orange, or green:
     - Blue: reachable leaf tensors that requires grad (tensors whose `.grad`
         fields will be populated during `.backward()`)
     - Orange: saved tensors of custom autograd functions as well as those
         saved by built-in backward nodes
     - Green: tensor passed in as outputs
     - Dark green: if any output is a view, we represent its base tensor with
         a dark green node.

    Args:
        module (nn.Module): PyTorch model
        *args (Tensor): input to the model
        additional_params (dict): dict of additional params to label nodes with
        show_attrs (bool): whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved (bool): whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars (int): if show_attrs is `True`, sets max number of characters
            to display for any given attribute.
        **kwargs (Tensor): input to the model

    Returns:
        AutoGradDot: graphviz representation of autograd graph
    """

    if LooseVersion(torch.__version__) < LooseVersion("1.9") and (show_attrs or show_saved):
        warnings.warn(
            "make_dot: showing grad_fn attributes and saved variables"
            " requires PyTorch version >= 1.9. (This does NOT apply to"
            " saved tensors saved by custom autograd functions.)"
        )

    autograd_dot = AutoGradDot()
    autograd_dot.module = module
    autograd_dot.inputs = args
    autograd_dot.inputs_kwargs = kwargs
    autograd_dot.reset_params()
    autograd_dot.show_attrs = show_attrs
    autograd_dot.show_saved = show_saved
    autograd_dot.max_attr_chars = max_attr_chars

    if additional_params is not None:
        autograd_dot.param_map.update(additional_params)

    # handle multiple outputs
    for v in autograd_dot.outputs:
        add_base_tensor(tensor=v, autograd_dot=autograd_dot)

    resize_graph(autograd_dot.dot)
    return autograd_dot


def join_scope_name(name: str, scope: Dict[str, str]) -> str:
    return '/'.join([scope[name], name])


def parse_trace_graph(graph) -> List[Node]:
    scope = {}
    for n in graph.nodes():
        inputs = [i.uniqueName() for i in n.inputs()]
        for i in range(1, len(inputs)):
            scope[inputs[i]] = n.scopeName()

        uname = next(n.outputs()).uniqueName()
        assert n.scopeName() != '', '{} has empty scope name'.format(n)
        scope[uname] = n.scopeName()

    scope['0'] = 'input'

    nodes = []
    for n in graph.nodes():
        attrs = {k: n[k] for k in n.attributeNames()}
        attrs = str(attrs).replace("'", ' ')
        inputs = [join_scope_name(i.uniqueName(), scope) for i in n.inputs()]
        uname = next(n.outputs()).uniqueName()

        nodes.append(Node(
            name=join_scope_name(uname, scope),
            op=n.kind(),
            inputs=inputs,
            attr=attrs
        ))

    for n in graph.inputs():
        uname = n.uniqueName()
        if uname not in scope.keys():
            scope[uname] = 'unused'

        nodes.append(Node(
            name=join_scope_name(uname, scope),
            op='Parameter',
            inputs=[],
            attr=str(n.type())
        ))

    return nodes


def get_autograd_dot_from_trace(trace) -> Digraph:
    """ Produces graphs of torch.jit.trace outputs

    Example:
    >>> trace, = torch.jit.trace(model, args=(x,))
    >>> dot = get_autograd_dot_from_trace(trace)

    Args:
        trace (torch.jit.trace): the trace object to visualize.

    Returns:
        graphviz.Digraph: the resulting graph.
    """

    try:
        from graphviz import Digraph
    except ImportError as e:
        raise ImportError("requires graphviz: https://pygraphviz.github.io/") from e

    # from tensorboardX
    if LooseVersion(torch.__version__) >= LooseVersion("0.4.1"):
        torch.onnx._optimize_trace(trace, torch._C._onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    elif LooseVersion(torch.__version__) >= LooseVersion("0.4"):
        torch.onnx._optimize_trace(trace, False)
    else:
        torch.onnx._optimize_trace(trace)
    graph = trace.graph()
    list_of_nodes = parse_trace_graph(graph)

    node_attr = dict(style='filled',
                     shape='box',
                     align='left',
                     fontsize='12',
                     ranksep='0.1',
                     height='0.2')

    dot = Digraph(node_attr=node_attr, graph_attr=dict(size="12,12"))

    for node in list_of_nodes:
        dot.node(node.name, label=node.name.replace('/', '\n'))
        if node.inputs:
            for inp in node.inputs:
                dot.edge(inp, node.name)

    resize_graph(dot)

    return dot


def get_autograd_obj(module: nn.Module, *args, **kwargs) -> AutoGradDot:
    """Produces Graphviz representation of PyTorch autograd graph.

    Args:
        module: PyTorch model
        *args: input to the make_autograd_obj
        **kwargs: input to the make_autograd_obj

    Returns:
        AutoGradDot: graphviz representation of autograd graph
    """
    assert isinstance(module, nn.Module)
    new_args = []
    new_kwargs = {}
    device = is_cpu_cuda.get_module_device(module)

    for i in args:
        assert isinstance(i, Tensor)
        i = i.to(device)
        i = i.detach()
        i = i.requires_grad_(True)
        new_args.append(i)

    for k, v in kwargs.items():
        assert isinstance(v, Tensor)
        v = v.to(device)
        v = v.detach()
        v = v.requires_grad_(True)
        new_kwargs[k] = v

    training_status = module.training
    use_autograd_graph = False
    if isinstance(module, Layer):
        use_autograd_graph = module.use_autograd_graph
        module.use_autograd_graph = True
    module.eval()

    autograd_dot = make_autograd_obj(module, *new_args, **new_kwargs)

    module.train(training_status)
    if isinstance(module, Layer):
        module.use_autograd_graph = use_autograd_graph
    return autograd_dot


def get_autograd_dot(module: nn.Module, *args, **kwargs) -> Digraph:
    """Produces Graphviz representation of PyTorch autograd graph.

    Args:
        module: PyTorch model
        *args: input to the make_autograd_obj
        **kwargs: input to the make_autograd_obj

    Returns:
        Digraph: graphviz representation of autograd graph
    """
    return get_autograd_obj(module, *args, **kwargs).dot


def save_autograd_graph(filename, module: nn.Module, *args, **kwargs) -> None:
    """Save Graphviz representation of PyTorch autograd graph.

    Args:
        filename: filename to save the graph to
        module: PyTorch model
        *args: input to the make_autograd_obj
        **kwargs: input to the make_autograd_obj
    """

    get_autograd_dot(module, *args, **kwargs).render(filename, format="svg", cleanup=True)


def save_autograd_graph_from_trace(filename, trace) -> None:
    """Save Graphviz representation of PyTorch autograd graph from trace.

    Args:
        filename: filename to save the graph to
        trace (torch.jit.trace): the trace object to visualize.
    """

    get_autograd_dot_from_trace(trace).render(filename, format="svg", cleanup=True)



from __future__ import annotations

import enum
from typing import Union, Callable

__all__ = ['GraphEnum', 'GRAPH_NODE_TYPE']


class GraphEnum(enum.Enum):
    """The graph enum for indicating input, output and stop.

    Attributes:
        INPUT (GraphEnum): GraphEnum.INPUT
        OUTPUT (GraphEnum): GraphEnum.OUTPUT
        STOP (GraphEnum): GraphEnum.STOP
    """
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"
    STOP = "STOP"


GRAPH_NODE_TYPE = Union[GraphEnum, Callable]



from __future__ import annotations

from typing import Optional

import torch

from analogvnn.nn.module.Sequential import Sequential

__all__ = ['FullSequential']


class FullSequential(Sequential):
    """A sequential model where backward graph is the reverse of forward graph.
    """

    def compile(self, device: Optional[torch.device] = None, layer_data: bool = True):
        """Compile the model and add forward and backward graph.

        Args:
            device (torch.device): The device to run the model on.
            layer_data (bool): True if the data of the layers should be compiled.

        Returns:
            FullSequential: self
        """
        arr = [self.graphs.INPUT, *list(self._runtime_module_list.values()), self.graphs.OUTPUT]
        self.graphs.forward_graph.add_connection(*arr)
        self.graphs.backward_graph.add_connection(*reversed(arr))
        return super().compile(device, layer_data)



from __future__ import annotations

import os
import re
from typing import Optional, Sequence

import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

from analogvnn.nn.module.Layer import Layer
from analogvnn.nn.module.Model import Model

__all__ = ['TensorboardModelLog']


class TensorboardModelLog:
    """Tensorboard model log.

    Attributes:
        model (nn.Module): the model to log.
        tensorboard (SummaryWriter): the tensorboard.
        layer_data (bool): whether to log the layer data.
    """
    model: nn.Module
    tensorboard: Optional[SummaryWriter]
    layer_data: bool

    def __init__(self, model: Model, log_dir: str):
        """Log the model to Tensorboard.

        Args:
            model (nn.Module): the model to log.
            log_dir (str): the directory to log to.
        """
        self.model = model
        self.tensorboard = None
        self.layer_data = True

        if not os.path.exists(log_dir):
            os.mkdir(log_dir)

        self.set_log_dir(log_dir)
        if hasattr(model, "subscribe_tensorboard"):
            model.subscribe_tensorboard(tensorboard=self)

    def set_log_dir(self, log_dir: str) -> TensorboardModelLog:
        """Set the log directory.

        Args:
            log_dir (str): the log directory.

        Returns:
            TensorboardModelLog: self.

        Raises:
            ValueError: if the log directory is invalid.
        """

        # https://github.com/tensorflow/tensorboard/pull/6135
        from tensorboard.compat import tf
        if getattr(tf, "io", None) is None:
            import tensorboard.compat.tensorflow_stub as new_tf
            tf.__dict__.update(new_tf.__dict__)

        if os.path.isdir(log_dir):
            self.tensorboard = SummaryWriter(log_dir=log_dir)
        else:
            raise ValueError(f"Log directory {log_dir} does not exist.")
        return self

    def _add_layer_data(self, epoch: int = None):
        """Add the layer data to the tensorboard.

        Args:
            epoch (int): the epoch to add the data for.
        """
        idx = 0
        for module in self.model.modules():
            if isinstance(module, nn.Sequential) or isinstance(module, nn.ModuleList) or (module == self):
                continue

            idx += 1
            if hasattr(module, "bias") and hasattr(module.bias, "size"):
                self.tensorboard.add_histogram(f"{idx}-{module}.bias", module.bias, epoch)
            if hasattr(module, "weight") and hasattr(module.weight, "size"):
                self.tensorboard.add_histogram(f"{idx}-{module}.weight", module.weight, epoch)

    def on_compile(self, layer_data: bool = True):
        """Called when the model is compiled.

        Args:
            layer_data (bool): whether to log the layer data.
        """
        if self.layer_data:
            self.layer_data = layer_data

        if self.layer_data:
            self._add_layer_data(epoch=-1)
        return self

    def add_graph(
            self,
            train_loader: DataLoader,
            model: Optional[nn.Module] = None,
            input_size: Optional[Sequence[int]] = None,
    ) -> TensorboardModelLog:
        """Add the model graph to the tensorboard.

        Args:
            train_loader (DataLoader): the train loader.
            model (Optional[nn.Module]): the model to log.
            input_size (Optional[Sequence[int]]): the input size.

        Returns:
            TensorboardModelLog: self.
        """
        if model is None:
            model = self.model

        log_id = f"{TensorboardModelLog.add_graph.__name__}_{id(model)}"
        if getattr(self.tensorboard, log_id, False):
            return self

        if input_size is None:
            data_shape = next(iter(train_loader))[0].shape
            input_size = [1] + list(data_shape)[1:]

        use_autograd_graph = False
        if isinstance(model, Layer):
            use_autograd_graph = model.use_autograd_graph
            model.use_autograd_graph = True

        self.tensorboard.add_graph(model, torch.zeros(input_size).to(model.device))
        setattr(self.tensorboard, log_id, True)

        if isinstance(model, Layer):
            model.use_autograd_graph = use_autograd_graph

        return self

    def add_summary(
            self,
            train_loader: DataLoader,
            model: Optional[nn.Module] = None,
            input_size: Optional[Sequence[int]] = None,
    ) -> TensorboardModelLog:
        """Add the model summary to the tensorboard.

        Args:
            train_loader (DataLoader): the train loader.
            model (nn.Module): the model to log.
            input_size (Optional[Sequence[int]]): the input size.

        Returns:
            TensorboardModelLog: self.

        Raises:
            ImportError: if torchinfo (https://github.com/tyleryep/torchinfo) is not installed.
        """

        try:
            import torchinfo
        except ImportError as e:
            raise ImportError("requires torchinfo: https://github.com/tyleryep/torchinfo") from e

        if model is None:
            model = self.model

        log_id = f"{TensorboardModelLog.add_summary.__name__}_{id(model)}"
        if getattr(self.tensorboard, log_id, False):
            return self

        if input_size is None:
            data_shape = next(iter(train_loader))[0].shape
            input_size = tuple(list(data_shape)[1:])

        use_autograd_graph = False
        if isinstance(model, Layer):
            use_autograd_graph = model.use_autograd_graph
            model.use_autograd_graph = True

        model_str = re.sub("\n", "\n    ", "    " + str(model))
        nn_model_summary = torchinfo.summary(
            model,
            input_size=input_size,
            verbose=torchinfo.Verbosity.QUIET,
            col_names=[e.value for e in torchinfo.ColumnSettings],
            depth=10,
        )
        nn_model_summary.formatting.verbose = torchinfo.Verbosity.VERBOSE
        nn_model_summary = re.sub("\n", "\n    ", f"    {nn_model_summary}")

        self.tensorboard.add_text(f"str ({model.__class__.__name__})", model_str)
        self.tensorboard.add_text(f"summary ({model.__class__.__name__})", nn_model_summary)
        setattr(self.tensorboard, log_id, True)

        if isinstance(model, Layer):
            model.use_autograd_graph = use_autograd_graph
        return self

    def register_training(self, epoch: int, train_loss: float, train_accuracy: float) -> TensorboardModelLog:
        """Register the training data.

        Args:
            epoch (int): the epoch.
            train_loss (float): the training loss.
            train_accuracy (float): the training accuracy.

        Returns:
            TensorboardModelLog: self.
        """
        self.tensorboard.add_scalar('Loss/train', train_loss, epoch)
        self.tensorboard.add_scalar("Accuracy/train", train_accuracy, epoch)
        if self.layer_data:
            self._add_layer_data(epoch=epoch)
        return self

    def register_testing(self, epoch: int, test_loss: float, test_accuracy: float) -> TensorboardModelLog:
        """Register the testing data.

        Args:
            epoch (int): the epoch.
            test_loss (float): the test loss.
            test_accuracy (float): the test accuracy.

        Returns:
            TensorboardModelLog: self.
        """
        self.tensorboard.add_scalar('Loss/test', test_loss, epoch)
        self.tensorboard.add_scalar("Accuracy/test", test_accuracy, epoch)
        return self

    # noinspection PyUnusedLocal
    def close(self, *args, **kwargs):
        """Close the tensorboard.
        """
        if self.tensorboard is not None:
            self.tensorboard.close()
            self.tensorboard = None

    __exit__ = close
    """Close the tensorboard."""
    __del__ = close
    """Close the tensorboard."""



from typing import Optional

import torch
from torch import Tensor

from analogvnn.nn.activation.Activation import Activation

__all__ = ['SiLU']


class SiLU(Activation):
    """Implements the SiLU activation function.
    """

    @staticmethod
    def forward(x: Tensor) -> Tensor:
        """forward pass of the SiLU

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return x / (1 + torch.exp(-x))

    def backward(self, grad_output: Optional[Tensor]) -> Optional[Tensor]:
        """backward pass of the SiLU

        Args:
            grad_output (Optional[Tensor]): the gradient of the output tensor.

        Returns:
            Optional[Tensor]: the gradient of the input tensor.
        """
        x = self.inputs
        neg_e = torch.exp(-x)
        grad = (1 + neg_e + x * neg_e) / torch.pow(1 + neg_e, 2)
        return grad_output * grad



from __future__ import annotations

from typing import Union, Dict, Optional

from torch import Tensor

from analogvnn.graph.ArgsKwargs import InputOutput, ArgsKwargs
from analogvnn.graph.GraphEnum import GraphEnum, GRAPH_NODE_TYPE

__all__ = ['ModelGraphState']


class ModelGraphState:
    """The state of a model graph.

    Attributes:
        allow_loops (bool): if True, the graph is allowed to contain loops.
        forward_input_output_graph (Optional[Dict[GRAPH_NODE_TYPE, InputOutput]]): the input and output of the
        forward pass.
        use_autograd_graph (bool): if True, the autograd graph is used to calculate the gradients.
        _loss (Tensor): the loss.
        INPUT (GraphEnum): GraphEnum.INPUT
        OUTPUT (GraphEnum): GraphEnum.OUTPUT
        STOP (GraphEnum): GraphEnum.STOP

    Properties:
        input (Tensor): the input of the forward pass.
        output (Tensor): the output of the forward pass.
        loss (Tensor): the loss.
    """
    allow_loops: bool
    use_autograd_graph: bool
    forward_input_output_graph: Optional[Dict[GRAPH_NODE_TYPE, InputOutput]]
    _loss: Optional[Tensor]

    INPUT = GraphEnum.INPUT
    OUTPUT = GraphEnum.OUTPUT
    STOP = GraphEnum.STOP

    def __init__(self, use_autograd_graph: bool = False, allow_loops=False):
        """Initialize the state.

        Args:
            use_autograd_graph: If True, the autograd graph is used to calculate the gradients.
            allow_loops: If True, the graph is allowed to contain loops.
        """
        self.allow_loops = allow_loops
        self.use_autograd_graph = use_autograd_graph

        self.forward_input_output_graph = None
        self._loss = None

    # noinspection PyUnusedLocal
    @staticmethod
    def ready_for_forward(exception: bool = False) -> bool:
        """Check if the state is ready for forward pass.

        Args:
            exception (bool): if True, raise an exception if the state is not ready for forward pass.

        Returns:
            bool: True if the state is ready for forward pass.
        """
        return True

    def ready_for_backward(self, exception: bool = False) -> bool:
        """Check if the state is ready for backward pass.

        Args:
            exception (bool): if True, raise an exception if the state is not ready for backward pass.

        Returns:
            bool: True if the state is ready for backward pass.

        Raises:
            RuntimeError: if the state is not ready for backward pass and exception is True.
        """
        if exception:
            if self.outputs is None:
                raise RuntimeError("output is not set.")

            if self._loss is None:
                raise RuntimeError("loss is not set.")
        else:
            return not (self.outputs is None or self._loss is None)

    @property
    def inputs(self) -> Optional[ArgsKwargs]:
        """Get the inputs.

        Returns:
            ArgsKwargs: the inputs.
        """
        if self.INPUT not in self.forward_input_output_graph:
            return None
        return self.forward_input_output_graph[self.INPUT].inputs

    @property
    def outputs(self) -> Optional[ArgsKwargs]:
        """Get the output.

        Returns:
            ArgsKwargs: the output.
        """
        if self.OUTPUT not in self.forward_input_output_graph:
            return None
        return self.forward_input_output_graph[self.OUTPUT].outputs

    @property
    def loss(self):
        """Get the loss.

        Returns:
            Tensor: the loss.
        """
        return self._loss

    def set_loss(self, loss: Union[Tensor, None]) -> ModelGraphState:
        """Set the loss.

        Args:
            loss (Tensor): the loss.

        Returns:
            ModelGraphState: self.
        """
        self._loss = loss
        return self



from __future__ import annotations

from typing import Dict, Sequence

import torch
from torch import Tensor

from analogvnn.graph.AcyclicDirectedGraph import AcyclicDirectedGraph
from analogvnn.graph.ArgsKwargs import ArgsKwargs, InputOutput, ArgsKwargsOutput
from analogvnn.graph.GraphEnum import GraphEnum
from analogvnn.utils.common_types import TENSORS

__all__ = ['ForwardGraph']


class ForwardGraph(AcyclicDirectedGraph):
    """The forward graph."""

    def __call__(self, inputs: TENSORS, is_training: bool) -> ArgsKwargsOutput:
        """Forward pass through the forward graph

        Args:
            inputs (TENSORS): Input to the graph
            is_training (bool): Is training or not

        Returns:
            ArgsKwargsOutput: Output of the graph
        """
        self.graph_state.ready_for_forward(exception=True)
        outputs = self.calculate(inputs, is_training)
        return outputs

    def compile(self, is_static: bool = True):
        """Compile the graph

        Args:
            is_static (bool): If True, the graph is not changing during runtime and will be cached.

        Returns:
            ForwardGraph: self.

        Raises:
            ValueError: If no forward pass has been performed yet.
        """
        if not self.graph.has_node(self.INPUT):
            raise ValueError("INPUT doesn't exist in the forward graph. Please preform a forward pass first.")

        if not self.graph.has_node(self.OUTPUT):
            raise ValueError("OUTPUT doesn't exist in the forward graph. Please preform a forward pass first.")

        return super().compile(is_static=is_static)

    def calculate(
            self,
            inputs: TENSORS,
            is_training: bool = True,
            **kwargs
    ) -> ArgsKwargsOutput:
        """Calculate the output of the graph

        Args:
            inputs (TENSORS): Input to the graph
            is_training (bool): Is training or not
            **kwargs: Additional arguments

        Returns:
            ArgsKwargsOutput: Output of the graph
        """
        if not isinstance(inputs, Sequence):
            inputs = (inputs,)

        if not self.graph_state.use_autograd_graph and is_training:
            for i in inputs:
                i.requires_grad = True

        input_output_graph = self._pass(self.INPUT, *inputs)
        if is_training:
            self.graph_state.forward_input_output_graph = input_output_graph

        outputs = input_output_graph[self.OUTPUT].outputs
        return ArgsKwargs.from_args_kwargs_object(outputs)

    def _pass(self, from_node: GraphEnum, *inputs: Tensor) -> Dict[GraphEnum, InputOutput]:
        """Perform the forward pass through the graph

        Args:
            from_node (GraphEnum): The node to  start the forward pass from
            *inputs (Tensor): Input to the graph

        Returns:
            Dict[GraphEnum, InputOutput]: The input and output of each node
        """
        static_graph = self._create_static_sub_graph(from_node)
        input_output_graph = {
            from_node: InputOutput(inputs=ArgsKwargs(args=[*inputs]))
        }
        for module, predecessors in static_graph:
            if module != from_node:
                inputs = self.parse_args_kwargs(input_output_graph, module, predecessors)
                if not self.graph_state.use_autograd_graph:
                    inputs.args = [self._detach_tensor(i) for i in inputs.args]
                    inputs.kwargs = {k: self._detach_tensor(v) for k, v in inputs.kwargs.items()}
                input_output_graph[module] = InputOutput(inputs=inputs)

            if isinstance(module, GraphEnum):
                input_output_graph[module].outputs = input_output_graph[module].inputs
                continue

            outputs = module(
                *input_output_graph[module].inputs.args,
                **input_output_graph[module].inputs.kwargs
            )
            input_output_graph[module].outputs = ArgsKwargs.to_args_kwargs_object(outputs)

        return input_output_graph

    @staticmethod
    def _detach_tensor(tensor: torch.Tensor) -> torch.Tensor:
        """Detach the tensor from the autograd graph

        Args:
            tensor (torch.Tensor): Tensor to detach

        Returns:
            torch.Tensor: Detached tensor
        """
        tensor: torch.Tensor = tensor.detach()
        tensor.requires_grad = True
        return tensor



import numpy as np

from analogvnn.utils.common_types import TENSOR_OPERABLE

__all__ = ['dirac_delta']


def dirac_delta(x: TENSOR_OPERABLE, a: TENSOR_OPERABLE = 0.001) -> TENSOR_OPERABLE:
    """`dirac_delta` takes `x` and returns the Dirac delta function of `x` with standard deviation of `a`.

    Args:
        x (TENSOR_OPERABLE): Tensor
        a (TENSOR_OPERABLE): standard deviation.

    Returns:
        TENSOR_OPERABLE: TENSOR_OPERABLE with the same shape as x, but with values equal to the Dirac delta function
        of x.
    """
    return 1 / (np.abs(a) * np.sqrt(np.pi)) * np.exp(-((x / a) ** 2))



from typing import Optional

from torch import Tensor

from analogvnn.nn.activation.Activation import Activation

__all__ = ['Identity']


class Identity(Activation):
    """Implements the identity activation function.

    Attributes:
        name (str): the name of the activation function.
    """
    name: Optional[str]

    def __init__(self, name=None):
        """initialize the identity activation function.

        Args:
            name (str): the name of the activation function.
        """
        super(Identity, self).__init__()
        self.name = name

    def extra_repr(self) -> str:
        """extra __repr__ of the identity activation function.

        Returns:
            str: the extra representation of the identity activation function.
        """
        if self.name is not None:
            return f'name={self.name}'
        else:
            return ''

    @staticmethod
    def forward(x: Tensor) -> Tensor:
        """forward pass of the identity activation function.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor same as the input tensor.
        """
        return x

    def backward(self, grad_output: Optional[Tensor]) -> Optional[Tensor]:
        """backward pass of the identity activation function.

        Args:
            grad_output (Optional[Tensor]): the gradient of the output tensor.

        Returns:
            Optional[Tensor]: the gradient of the input tensor same as the gradient of the output tensor.
        """
        return grad_output



from __future__ import annotations

from typing import Callable, Any

import torch
import torch.nn as nn
from torch import Tensor
from torch.nn import ModuleDict
from torch.nn.utils import parametrize

from analogvnn.parameter.Parameter import Parameter

__all__ = ['PseudoParameter']


class PseudoParameterModule(nn.Module):
    """A module that wraps a parameter and a function to transform it.

    Attributes:
        original (PseudoParameter): the original parameters.
        _transformed (nn.Parameter): the transformed parameters.
    """
    original: PseudoParameter
    _transformed: nn.Parameter

    def __init__(self, original, transformed):
        """Creates a new pseudo parameter module.

        Args:
            original (PseudoParameter): the original parameters.
            transformed (nn.Parameter): the transformed parameters.
        """
        super().__init__()
        self.original = original
        self._transformed = transformed

    def __call__(self, *args, **kwargs) -> nn.Parameter:
        """Transforms the parameter by calling the __call__ method of the PseudoParameter.

        Args:
            *args: additional arguments.
            **kwargs: additional keyword arguments.

        Returns:
            nn.Parameter: The transformed parameter.
        """
        return self.original()

    forward = __call__
    """Alias for __call__"""

    _call_impl = __call__
    """Alias for __call__"""

    def set_original_data(self, data: Tensor) -> PseudoParameterModule:
        """set data to the original parameter.

        Args:
            data (Tensor): the data to set.

        Returns:
            PseudoParameterModule: self.
        """
        self.original.data = data
        return self

    right_inverse = set_original_data
    """Alias for set_original_data."""


class PseudoParameter(Parameter):
    """A parameterized parameter which acts like a normal parameter during gradient updates.

    PyTorch's ParameterizedParameters vs AnalogVNN's PseudoParameters:

    - Similarity (Forward or Parameterizing the data):
        > Data -> ParameterizingModel -> Parameterized Data

    - Difference (Backward or Gradient Calculations):
      - ParameterizedParameters
        > Parameterized Data -> ParameterizingModel -> Data
      - PseudoParameters
        > Parameterized Data -> Data

    Attributes:
        _transformation (Callable): the transformation.
        _transformed (nn.Parameter): the transformed parameter.
        _module (PseudoParameterModule): the module that wraps the parameter and the transformation.

    Properties:
        grad (Tensor): the gradient of the parameter.
        module (PseudoParameterModule): the module that wraps the parameter and the transformation.
        transformation (Callable): the transformation.
    """
    _transformation: Callable
    _transformed: nn.Parameter
    _module: PseudoParameterModule

    @staticmethod
    def identity(x: Any) -> Any:
        """The identity function.

        Args:
            x (Any): the input tensor.

        Returns:
            Any: the input tensor.
        """
        return x

    def __init__(self, data=None, requires_grad=True, transformation=None, *args, **kwargs):
        """Initializes the parameter.

        Args:
            data: the data for the parameter.
            requires_grad (bool): whether the parameter requires gradient.
            transformation (Callable): the transformation.
            *args: additional arguments.
            **kwargs: additional keyword arguments.
        """
        super().__init__(data, requires_grad, *args, **kwargs)
        self._transformed = nn.Parameter(data=data, requires_grad=requires_grad)
        self._transformed.original = self
        self._transformation = self.identity
        self.set_transformation(transformation)

        self._module = PseudoParameterModule(
            original=self,
            transformed=self._transformed
        )

    def __call__(self, *args, **kwargs):
        """Transforms the parameter.

        Args:
            *args: additional arguments.
            **kwargs: additional keyword arguments.

        Returns:
            nn.Parameter: the transformed parameter.

        Raises:
            RuntimeError: if the transformation callable fails.
        """
        try:
            self._transformed.data = self._transformation(self)
        except Exception as e:
            raise RuntimeError(f"here: {e.args}") from e
        return self._transformed

    def __repr__(self):
        """Returns a string representation of the parameter.

        Returns:
            str: the string representation.
        """
        return f'{PseudoParameter.__name__}(' \
               f'transform={self.transformation}' \
               f', data={self.data}' \
               f')'

    @property
    def grad(self):
        """Returns the gradient of the parameter.

        Returns:
            Tensor: the gradient.
        """
        return self._transformed.grad

    @property
    def module(self):
        """Returns the module.

        Returns:
            PseudoParameterModule: the module.
        """
        return self._module

    @property
    def transformation(self):
        """Returns the transformation.

        Returns:
            Callable: the transformation.
        """
        return self._transformation

    @transformation.setter
    def transformation(self, transformation: Callable):
        """Sets the transformation.

        Args:
            transformation (Callable): the transformation.
        """
        self.set_transformation(transformation)

    def set_transformation(self, transformation) -> PseudoParameter:
        """Sets the transformation.

        Args:
            transformation (Callable): the transformation.

        Returns:
            PseudoParameter: self.
        """
        self._transformation = transformation
        if isinstance(self._transformation, nn.Module):
            self._transformation.eval()
        return self

    @classmethod
    def parameterize(cls, module: nn.Module, param_name: str, transformation: Callable) -> PseudoParameter:
        """Parameterizes a parameter.

        Args:
            module (nn.Module): the module.
            param_name (str): the name of the parameter.
            transformation (Callable): the transformation to apply.

        Returns:
            PseudoParameter: the parameterized parameter.
        """
        assert hasattr(module, param_name)

        param = getattr(module, param_name)
        new_param = cls(data=param, requires_grad=param.requires_grad, transformation=transformation)

        delattr(module, param_name)

        if not parametrize.is_parametrized(module):
            # Change the class
            parametrize._inject_new_class(module)
            # Inject a ``ModuleDict`` into the instance under module.parametrizations
            module.parametrizations = ModuleDict()

        module.parametrizations[param_name] = new_param.module
        parametrize._inject_property(module, param_name)
        return new_param

    @classmethod
    def parametrize_module(cls, module: nn.Module, transformation: Callable, requires_grad: bool = True):
        """Parametrize all parameters of a module.

        Args:
            module (nn.Module): the module parameters to parametrize.
            transformation (Callable): the transformation.
            requires_grad (bool): if True, only parametrized parameters that require gradients.
        """
        with torch.no_grad():
            for name, parameter in list(module.named_parameters(recurse=False)):
                if isinstance(parameter, cls):
                    continue

                if requires_grad and not parameter.requires_grad:
                    continue

                cls.parameterize(module=module, param_name=name, transformation=transformation)

            for sub_module in module.children():
                if sub_module == module:
                    continue
                if hasattr(module, "parametrizations") and (
                        sub_module is module.parametrizations or sub_module in module.parametrizations
                ):
                    continue
                cls.parametrize_module(module=sub_module, transformation=transformation, requires_grad=requires_grad)



from __future__ import annotations

from analogvnn.graph.BackwardGraph import BackwardGraph
from analogvnn.graph.ForwardGraph import ForwardGraph
from analogvnn.graph.ModelGraphState import ModelGraphState

__all__ = ['ModelGraph']


class ModelGraph(ModelGraphState):
    """Store model's graph.

    Attributes:
        forward_graph (ForwardGraph): store model's forward graph.
        backward_graph (BackwardGraph): store model's backward graph.
    """
    forward_graph: ForwardGraph
    backward_graph: BackwardGraph

    def __init__(self, use_autograd_graph: bool = False, allow_loops: bool = False):
        """Initialize the model graph.

        Args:
            use_autograd_graph: If True, the autograd graph is used to calculate the gradients.
            allow_loops: If True, the graph is allowed to contain loops.
        """
        super(ModelGraph, self).__init__(use_autograd_graph, allow_loops)
        self.forward_graph = ForwardGraph(self)
        self.backward_graph = BackwardGraph(self)

    def compile(self, is_static: bool = True, auto_backward_graph: bool = False) -> ModelGraph:
        """Compile the model graph.

        Args:
            is_static (bool): If True, the model graph is static.
            auto_backward_graph (bool): If True, the backward graph is automatically created.

        Returns:
            ModelGraph: self.
        """
        self.forward_graph.compile(is_static=is_static)

        if auto_backward_graph:
            self.backward_graph.from_forward(self.forward_graph)

        self.backward_graph.compile(is_static=is_static)
        return self




import torch
from torch import nn, Tensor

from analogvnn.nn.normalize.Normalize import Normalize

__all__ = ['LPNorm', 'LPNormW', 'L1Norm', 'L2Norm', 'L1NormW', 'L2NormW', 'L1NormM', 'L2NormM', 'L1NormWM', 'L2NormWM']


class LPNorm(Normalize):
    """Implements the row-wise Lp normalization function.

    Attributes:
        p (int): the pth power of the Lp norm.
        make_max_1 (bool): if True, the maximum absolute value of the output tensor will be 1.
    """
    __constants__ = ['p', 'make_max_1']
    p: nn.Parameter
    make_max_1: nn.Parameter

    def __init__(self, p: int, make_max_1=False):
        """initializes the row-wise Lp normalization function.

        Args:
            p (int): the pth power of the Lp norm.
            make_max_1 (bool): if True, the maximum absolute value of the output tensor will be 1.
        """
        super(LPNorm, self).__init__()
        self.p = nn.Parameter(torch.tensor(p), requires_grad=False)
        self.make_max_1 = nn.Parameter(torch.tensor(make_max_1), requires_grad=False)

    def forward(self, x: Tensor):
        """forward pass of row-wise Lp normalization function.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        norm = x
        if len(x.shape) > 1:
            norm = torch.flatten(norm, start_dim=1)

        norm = torch.norm(norm, self.p, -1)
        norm = torch.clamp(norm, min=1e-4)
        x = torch.div(x.T, norm).T

        if self.make_max_1:
            x = torch.div(x, torch.max(torch.abs(x)))

        return x


class LPNormW(LPNorm):
    """Implements the whole matrix Lp normalization function.
    """

    def forward(self, x: Tensor):
        norm = torch.norm(x, self.p)
        norm = torch.clamp(norm, min=1e-4)
        x = torch.div(x, norm)

        if self.make_max_1:
            x = torch.div(x, torch.max(torch.abs(x)))

        return x


class L1Norm(LPNorm):
    """Implements the row-wise L1 normalization function.
    """

    def __init__(self):
        super(L1Norm, self).__init__(p=1, make_max_1=False)


class L2Norm(LPNorm):
    """Implements the row-wise L2 normalization function.
    """

    def __init__(self):
        super(L2Norm, self).__init__(p=2, make_max_1=False)


class L1NormW(LPNormW):
    """Implements the whole matrix L1 normalization function.
    """

    def __init__(self):
        super(L1NormW, self).__init__(p=1, make_max_1=False)


class L2NormW(LPNormW):
    """Implements the whole matrix L2 normalization function.
    """

    def __init__(self):
        super(L2NormW, self).__init__(p=2, make_max_1=False)


class L1NormM(LPNorm):
    """Implements the row-wise L1 normalization function with maximum absolute value of 1.
    """

    def __init__(self):
        super(L1NormM, self).__init__(p=1, make_max_1=True)


class L2NormM(LPNorm):
    """Implements the row-wise L2 normalization function with maximum absolute value of 1.
    """

    def __init__(self):
        super(L2NormM, self).__init__(p=2, make_max_1=True)


class L1NormWM(LPNormW):
    """Implements the whole matrix L1 normalization function with maximum absolute value of 1.
    """

    def __init__(self):
        super(L1NormWM, self).__init__(p=1, make_max_1=True)


class L2NormWM(LPNormW):
    """Implements the whole matrix L2 normalization function with maximum absolute value of 1.
    """

    def __init__(self):
        super(L2NormWM, self).__init__(p=2, make_max_1=True)



from __future__ import annotations

import abc
import typing
from typing import Dict, Union, List, Tuple

import networkx as nx

from analogvnn.graph.ArgsKwargs import ArgsKwargs, InputOutput
from analogvnn.graph.GraphEnum import GraphEnum, GRAPH_NODE_TYPE
from analogvnn.graph.ModelGraphState import ModelGraphState
from analogvnn.graph.to_graph_viz_digraph import to_graphviz_digraph

if typing.TYPE_CHECKING:
    pass

__all__ = ['AcyclicDirectedGraph']


class AcyclicDirectedGraph(abc.ABC):
    """The base class for all acyclic directed graphs.

    Attributes:
        graph (nx.MultiDiGraph): The graph.
        graph_state (ModelGraphState): The graph state.
        _is_static (bool): If True, the graph is not changing during runtime and will be cached.
        _static_graphs (Dict[GRAPH_NODE_TYPE, List[Tuple[GRAPH_NODE_TYPE, List[GRAPH_NODE_TYPE]]]]): The static graphs.
        INPUT (GraphEnum): GraphEnum.INPUT
        OUTPUT (GraphEnum): GraphEnum.OUTPUT
        STOP (GraphEnum): GraphEnum.STOP

    """
    graph: nx.MultiDiGraph
    graph_state: ModelGraphState
    _is_static: bool
    _static_graphs: Dict[GRAPH_NODE_TYPE, List[Tuple[GRAPH_NODE_TYPE, List[GRAPH_NODE_TYPE]]]]

    INPUT = GraphEnum.INPUT
    OUTPUT = GraphEnum.OUTPUT
    STOP = GraphEnum.STOP

    def __init__(self, graph_state: ModelGraphState = None):
        """Create a new graph.

        Args:
            graph_state (ModelGraphState): The graph state.

        Raises:
            NotImplementedError: If allow_loops is True, since this is not implemented yet.
        """
        self.graph = nx.MultiDiGraph()
        self.graph_state = graph_state
        self._is_static = False
        self._static_graphs = {}

        if self.graph_state.allow_loops:
            raise NotImplementedError("Loops are not implemented yet. Coming soon...")

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        """Performs pass through the graph

        Args:
            *args: Arguments
            **kwargs: Keyword arguments

        Raises:
            NotImplementedError: since method is abstract
        """
        raise NotImplementedError

    def add_connection(self, *args: GRAPH_NODE_TYPE):
        """Add a connection between nodes.

        Args:
            *args: The nodes.

        Returns:
            AcyclicDirectedGraph: self.
        """
        for i in range(1, len(args)):
            self.add_edge(args[i - 1], args[i])
        return self

    def add_edge(
            self,
            u_of_edge: GRAPH_NODE_TYPE,
            v_of_edge: GRAPH_NODE_TYPE,
            in_arg: Union[None, int, bool] = None,
            in_kwarg: Union[None, str, bool] = None,
            out_arg: Union[None, int, bool] = None,
            out_kwarg: Union[None, str, bool] = None
    ):
        """Add an edge to the graph.

        Args:
            u_of_edge (GRAPH_NODE_TYPE): The source node.
            v_of_edge (GRAPH_NODE_TYPE): The target node.
            in_arg (Union[None, int, bool]): The input argument.
            in_kwarg (Union[None, str, bool]): The input keyword argument.
            out_arg (Union[None, int, bool]): The output argument.
            out_kwarg (Union[None, str, bool]): The output keyword argument.

        Returns:
            AcyclicDirectedGraph: self.
        """
        attr = self.check_edge_parameters(in_arg, in_kwarg, out_arg, out_kwarg)
        existing_edges = self.graph.get_edge_data(u_of_edge, v_of_edge)

        if existing_edges is not None:
            to_remove = []
            for key, edge in existing_edges.items():
                if not (
                        edge["out_arg"] == attr["out_arg"] is not None
                        or
                        edge["out_kwarg"] == attr["out_kwarg"] is not None
                ):
                    continue
                to_remove.append(key)
            for key in to_remove:
                self.graph.remove_edge(u_of_edge, v_of_edge, key=key)

        self.graph.add_edge(u_of_edge, v_of_edge, **attr)
        self.graph.nodes[u_of_edge]["fillcolor"] = "lightblue"
        self.graph.nodes[v_of_edge]["fillcolor"] = "lightblue"
        return self

    @staticmethod
    def check_edge_parameters(
            in_arg: Union[None, int, bool],
            in_kwarg: Union[None, str, bool],
            out_arg: Union[None, int, bool],
            out_kwarg: Union[None, str, bool]
    ) -> Dict[str, Union[None, int, str, bool]]:
        """Check the edge's in and out parameters.

        Args:
            in_arg (Union[None, int, bool]): The input argument.
            in_kwarg (Union[None, str, bool]): The input keyword argument.
            out_arg (Union[None, int, bool]): The output argument.
            out_kwarg (Union[None, str, bool]): The output keyword argument.

        Returns:
            Dict[str, Union[None, int, str, bool]]: Dict of valid edge's in and out parameters.

        Raises:
            ValueError: If in and out parameters are invalid.
        """
        # @@@ in_arg: None    in_kwarg: None  out_arg: None   out_kwarg: None   0
        # @@  in_arg: True    in_kwarg: None  out_arg: True   out_kwarg: None   1
        #     in_arg: None    in_kwarg: True  out_arg: True   out_kwarg: None   2
        # @   in_arg: None    in_kwarg: True  out_arg: None   out_kwarg: True   3
        # @   in_arg: 0       in_kwarg: None  out_arg: True   out_kwarg: None   4
        #     in_arg: 0       in_kwarg: None  out_arg: 0      out_kwarg: None   5
        #     in_arg: 0       in_kwarg: None  out_arg: None   out_kwarg: a      6
        #     in_arg: None    in_kwarg: a     out_arg: True   out_kwarg: None   7
        #     in_arg: None    in_kwarg: a     out_arg: 0      out_kwarg: None   8
        # @@  in_arg: None    in_kwarg: a     out_arg: None   out_kwarg: a      9
        if out_arg is not None and out_kwarg is not None:
            raise ValueError('both "out_arg" and "out_kwarg" can\'t be present at the same time')
        if in_arg is not None and in_kwarg is not None:
            raise ValueError('both "in_arg" and "in_kwarg" can\'t be present at the same time')

        if in_arg is None and in_kwarg is None and (out_arg is True or out_kwarg is True):
            in_arg = out_arg
            in_kwarg = out_kwarg
        if in_arg is None and in_kwarg is None and (out_arg is not None or out_kwarg is not None):
            in_arg = 0

        if in_arg is True or in_kwarg is True:
            #  All -> All
            if in_arg not in [True, None]:
                raise ValueError(f'Invalid value for in_arg: "{in_arg}')
            if in_kwarg not in [True, None]:
                raise ValueError(f'Invalid value for in_kwarg: "{in_kwarg}')
            if out_arg not in [True, None]:
                raise ValueError(f'Invalid value for out_arg: "{out_arg}')
            if out_kwarg not in [True, None]:
                raise ValueError(f'Invalid value for out_kwarg: "{out_kwarg}')

            if in_arg is True and out_kwarg is True:
                raise ValueError(f'Invalid value in_arg -> out_kwarg')

            if (in_arg is True or in_kwarg is True) and (out_arg is None and out_kwarg is None):
                out_arg = in_arg
                out_kwarg = in_kwarg
        elif in_arg is not None or in_kwarg is not None:
            # one -> one
            if in_arg is not None and (not isinstance(in_arg, int) or in_arg < 0):
                raise ValueError('"in_arg" must be a number >= 0')
            if out_arg is not None and (not isinstance(out_arg, int) or out_arg < 0):
                raise ValueError('"out_arg" must be a number >= 0')
            if in_kwarg is not None and not isinstance(in_kwarg, str):
                raise ValueError('"in_kwarg" must be a string')
            if out_kwarg is not None and not isinstance(out_kwarg, str):
                raise ValueError('"out_kwarg" must be a string')

            if in_arg is not None and (out_arg is None and out_kwarg is None):
                out_arg = True
            if in_kwarg is not None and (out_arg is None and out_kwarg is None):
                out_kwarg = in_kwarg
        else:
            in_arg = True
            out_arg = True
            in_kwarg = True
            out_kwarg = True

        attr = {
            "in_arg": in_arg,
            "in_kwarg": in_kwarg,
            "out_arg": out_arg,
            "out_kwarg": out_kwarg,
        }
        attr["label"] = AcyclicDirectedGraph._create_edge_label(**attr)

        return attr

    # noinspection PyUnusedLocal
    @staticmethod
    def _create_edge_label(
            in_arg: Union[None, int, bool] = None,
            in_kwarg: Union[None, str, bool] = None,
            out_arg: Union[None, int, bool] = None,
            out_kwarg: Union[None, str, bool] = None,
            **kwargs
    ) -> str:
        """Create the edge's label.

        Args:
            in_arg (Union[None, int, bool]): The input argument.
            in_kwarg (Union[None, str, bool]): The input keyword argument.
            out_arg (Union[None, int, bool]): The output argument.
            out_kwarg (Union[None, str, bool]): The output keyword argument.

        Returns:
            str: The edge's label.
        """
        label = ""
        if in_arg == in_kwarg == out_arg == out_kwarg is True:
            return "* -> *"

        if in_arg is True:
            label += "[]"
        elif in_arg is not None:
            label += "[" + str(in_arg) + "]"
        if in_kwarg is True:
            label += "{}"
        elif in_kwarg is not None:
            label += "{" + str(in_kwarg) + "}"

        label += " -> "
        if out_arg is True:
            label += "[]"
        elif out_arg is not None:
            label += "[" + str(out_arg) + "]"
        if out_kwarg is True:
            label += "{}"
        elif out_kwarg is not None:
            label += "{" + str(out_kwarg) + "}"

        return label

    def compile(self, is_static: bool = True):
        """Compile the graph.

        Args:
            is_static (bool): If True, the graph will be compiled as a static graph.

        Returns:
            AcyclicDirectedGraph: The compiled graph.

        Raises:
            ValueError: If the graph is not acyclic.
        """
        for i in nx.simple_cycles(self.graph):
            raise ValueError(f'Cycle detected: {i}')

        self.graph = self._reindex_out_args(self.graph)
        self._is_static = is_static
        self._static_graphs = {}
        return self

    @staticmethod
    def _reindex_out_args(graph: nx.MultiDiGraph) -> nx.MultiDiGraph:
        """Reindex the output arguments.

        Args:
            graph (nx.MultiDiGraph): The graph.

        Returns:
            nx.MultiDiGraph: The graph with re-indexed output arguments.
        """
        # noinspection PyTypeChecker
        graph: nx.MultiDiGraph = graph.copy()

        for v in graph.nodes():
            args_index = []
            for u in graph.predecessors(v):
                for _, edge_data in graph.get_edge_data(u, v).items():
                    if isinstance(edge_data["out_arg"], int) and not isinstance(edge_data["out_arg"], bool):
                        args_index.append(edge_data)

            if len(args_index) == 0:
                continue

            args_index = sorted(args_index, key=lambda x: x["out_arg"])
            for index, value in enumerate(args_index):
                value["out_arg"] = index
                value["label"] = AcyclicDirectedGraph._create_edge_label(**value)

        return graph

    def _create_static_sub_graph(
            self,
            from_node: GRAPH_NODE_TYPE
    ) -> List[Tuple[GRAPH_NODE_TYPE, List[GRAPH_NODE_TYPE]]]:
        """Create a static sub graph connected to the given node.

        Args:
            from_node (GRAPH_NODE_TYPE): The node.

        Returns:
            List[Tuple[GRAPH_NODE_TYPE, List[GRAPH_NODE_TYPE]]]: The static sub graph.
        """
        if self._is_static and from_node in self._static_graphs:
            return self._static_graphs[from_node]

        nodes = nx.descendants(self.graph, from_node)
        nodes.add(from_node)

        sub_graph: nx.MultiDiGraph = self.graph.subgraph(nodes)
        sorted_graph = nx.topological_sort(sub_graph)
        dependent_sorted_graph = []

        for i in sorted_graph:
            dependent_sorted_graph.append((i, list(sub_graph.predecessors(i))))

        if self._is_static:
            self._static_graphs[from_node] = dependent_sorted_graph

        return self._static_graphs[from_node]

    def parse_args_kwargs(
            self,
            input_output_graph: Dict[GRAPH_NODE_TYPE, InputOutput],
            module: GRAPH_NODE_TYPE,
            predecessors: List[GRAPH_NODE_TYPE]
    ) -> ArgsKwargs:
        """Parse the arguments and keyword arguments.

        Args:
            input_output_graph (Dict[GRAPH_NODE_TYPE, InputOutput]): The input output graph.
            module (GRAPH_NODE_TYPE): The module.
            predecessors (List[GRAPH_NODE_TYPE]): The predecessors.

        Returns:
            ArgsKwargs: The arguments and keyword arguments.
        """
        args = {}
        extra_args = []
        kwargs = {}
        for p in predecessors:
            edge_data = self.graph.get_edge_data(p, module)
            for k, v in edge_data.items():
                previous_outputs = input_output_graph[p].outputs
                in_arg = v["in_arg"]
                in_kwarg = v["in_kwarg"]
                out_arg = v["out_arg"]
                out_kwarg = v["out_kwarg"]

                # 0
                if in_arg is True and out_arg is True and in_kwarg is True and out_kwarg is True:
                    extra_args += previous_outputs.args
                    kwargs.update(previous_outputs.kwargs)
                    continue
                # 1
                if in_arg is True and out_arg is True:
                    extra_args += previous_outputs.args
                    continue
                # 3
                if in_kwarg is True and out_kwarg is True:
                    kwargs.update(previous_outputs.kwargs)
                    continue
                # 2
                if in_kwarg is True and out_arg is True:
                    extra_args += list(previous_outputs.kwargs.values())
                    continue
                # Backward - 0
                if in_arg is True and out_kwarg is not None:
                    kwargs[out_kwarg] = previous_outputs.args
                    continue
                # Backward - 0
                if in_kwarg is True and out_kwarg is not None:
                    kwargs[out_kwarg] = previous_outputs.kwargs
                    continue

                # 4, 5 & 6
                if in_arg is not None:
                    previous_outputs = previous_outputs.args[in_arg]
                # 7, 8 & 9
                if in_kwarg is not None:
                    previous_outputs = previous_outputs.kwargs[in_kwarg]

                # 4 & 7
                if out_arg is True:
                    extra_args.append(previous_outputs)
                    continue
                # 5 & 8
                if out_arg is not None:
                    args[out_arg] = previous_outputs
                    continue
                # 6 & 9
                if out_kwarg is not None:
                    kwargs[out_kwarg] = previous_outputs
                    continue

                raise NotImplementedError("WTF!Why!")

        args = [args[k] for k in sorted(args.keys())] + extra_args
        inputs = ArgsKwargs(
            args=args,
            kwargs=kwargs
        )
        return inputs

    def render(self, *args, real_label: bool = False, **kwargs) -> str:
        """Save the source to file and render with the Graphviz engine.

        Args:
            *args: Arguments to pass to graphviz render function.
            real_label: If True, the real label will be used instead of the label.
            **kwargs: Keyword arguments to pass to graphviz render function.

        Returns:
            str: The (possibly relative) path of the rendered file.
        """
        return to_graphviz_digraph(self.graph, real_label=real_label).render(*args, **kwargs)



from typing import Dict, Union, Callable

import torch
from torch import nn

from analogvnn.graph.ArgsKwargs import ArgsKwargs, InputOutput
from analogvnn.graph.GraphEnum import GRAPH_NODE_TYPE

__all__ = ['AccumulateGrad']


class AccumulateGrad:
    """AccumulateGrad is a module that accumulates the gradients of the outputs of the module
    it is attached to. It has no parameters of its own.

    Attributes:
        module (nn.Module): Module to accumulate gradients for.
        input_output_connections (Dict[str, Dict[str, Union[None, bool, int, str, GRAPH_NODE_TYPE]]]): input/output
        connections.
    """
    input_output_connections: Dict[str, Dict[str, Union[None, bool, int, str, GRAPH_NODE_TYPE]]]
    module: Union[nn.Module, Callable]

    def __init__(self, module: Union[nn.Module, Callable]):
        """Initialize the module.

        Args:
            module (Union[nn.Module, Callable]): Module from which to accumulate gradients.
        """
        self.input_output_connections = {}
        self.module = module

    def __repr__(self):
        """Return a string representation of the module.

        Returns:
            str: String representation of the module.
        """
        return f"AccumulateGrad({self.module})"

    def __call__(
            self,
            grad_outputs_args_kwargs: ArgsKwargs,
            forward_input_output_graph: Dict[GRAPH_NODE_TYPE, InputOutput]
    ) -> ArgsKwargs:
        """Calculate and Accumulate the output gradients of the module.

        Args:
            grad_outputs_args_kwargs (ArgsKwargs): The output gradients from previous modules (predecessors).
            forward_input_output_graph (Dict[GRAPH_NODE_TYPE, InputOutput]): The input and output from forward pass.

        Returns:
            ArgsKwargs: The output gradients.
        """
        grad_inputs_args = {}
        grad_inputs_kwargs = {}
        for key, grad_output in grad_outputs_args_kwargs.kwargs.items():
            location = self.input_output_connections[key]
            forward_out_arg: Union[None, int, bool] = location['in_arg']
            forward_out_kwarg: Union[None, str, bool] = location['in_kwarg']
            forward_in_arg: Union[None, int, bool] = location['out_arg']
            forward_in_kwarg: Union[None, str, bool] = location['out_kwarg']
            predecessor: GRAPH_NODE_TYPE = location['from']

            # 0 - not allowed

            # 4
            if forward_out_arg is True and isinstance(forward_in_arg, int) and not isinstance(forward_in_arg, bool):
                forward_inputs = forward_input_output_graph[predecessor].inputs.args
                forward_outputs = forward_input_output_graph[self.module].outputs.args
                forward_out_arg = forward_inputs.index(forward_outputs[forward_in_arg])
                grad_output = grad_output[forward_out_arg]

            # 7
            if forward_out_arg is True and isinstance(forward_in_kwarg, str):
                forward_inputs = forward_input_output_graph[predecessor].inputs.args
                forward_outputs = forward_input_output_graph[self.module].outputs.kwargs
                forward_out_arg = forward_inputs.index(forward_outputs[forward_in_kwarg])
                grad_output = grad_output[forward_out_arg]

            # 1
            if forward_out_arg is True and forward_in_arg is True:
                forward_inputs = forward_input_output_graph[predecessor].inputs.args
                forward_outputs = forward_input_output_graph[self.module].outputs.args
                for i in range(len(forward_inputs)):
                    if forward_inputs[i] not in forward_outputs:
                        continue

                    value_index = forward_outputs.index(forward_inputs[i])
                    if value_index not in grad_inputs_args:
                        grad_inputs_args[value_index] = torch.zeros_like(grad_output[i])
                    grad_inputs_args[value_index] += grad_output[i]
                continue

            # 2
            if forward_out_arg is True and forward_in_kwarg is True:
                forward_inputs = forward_input_output_graph[predecessor].inputs.args
                forward_outputs = forward_input_output_graph[self.module].outputs.kwargs
                for i in forward_outputs:
                    value_index = forward_inputs.index(forward_outputs[i])

                    if i not in grad_inputs_kwargs:
                        grad_inputs_kwargs[i] = torch.zeros_like(grad_output[value_index])
                    grad_inputs_kwargs[i] += grad_output[value_index]
                continue

            # 3
            if forward_out_kwarg is True and forward_in_kwarg is True:
                for i in grad_output:
                    if i not in grad_inputs_kwargs:
                        grad_inputs_kwargs[i] = torch.zeros_like(grad_output[i])

                    grad_inputs_kwargs[i] += grad_output[i]
                continue

            # 8 & 9
            if forward_in_kwarg is not None and isinstance(forward_in_kwarg, str):
                if forward_in_kwarg not in grad_inputs_kwargs:
                    grad_inputs_kwargs[forward_in_kwarg] = torch.zeros_like(grad_output)

                grad_inputs_kwargs[forward_in_kwarg] += grad_output
                continue

            # 5 & 6
            if forward_in_arg is not None and isinstance(forward_in_arg, int) and not isinstance(forward_in_arg, bool):
                if forward_in_arg not in grad_inputs_args:
                    grad_inputs_args[forward_in_arg] = torch.zeros_like(grad_output)
                grad_inputs_args[forward_in_arg] += grad_output
                continue

            raise NotImplementedError("WTF!Why!")

        return ArgsKwargs(
            args=[grad_inputs_args[i] for i in sorted(list(grad_inputs_args.keys()))],
            kwargs=grad_inputs_kwargs
        )

    grad = __call__
    """Alias for __call__."""



from __future__ import annotations

import operator
from collections import OrderedDict
from itertools import islice
from typing import Iterator, TypeVar, Union, Dict, Optional

import torch
from torch import nn
from torch.nn import Module

from analogvnn.nn.module.Model import Model

T = TypeVar('T', bound=nn.Module)

__all__ = ['Sequential']


class Sequential(Model):
    """A sequential model.

    Attributes:
        _runtime_module_list (OrderedDict[str, nn.Module]): The ordered dictionary of the modules.
    """

    def __init__(self, *args):
        super(Sequential, self).__init__()
        self._runtime_module_list: Dict[str, Optional[Module]] = OrderedDict()
        self.add_sequence(*args)

    def compile(self, device: Optional[torch.device] = None, layer_data: bool = True):
        """Compile the model and add forward graph.

        Args:
            device (torch.device): The device to run the model on.
            layer_data (bool): True if the data of the layers should be compiled.

        Returns:
            Sequential: self
        """
        arr = [self.graphs.INPUT, *list(self._runtime_module_list.values()), self.graphs.OUTPUT]
        self.graphs.forward_graph.add_connection(*arr)
        return super().compile(device, layer_data)

    def add_sequence(self, *args):
        """Add a sequence of modules to the forward graph of model.

        Args:
            *args (nn.Module): The modules to add.
        """
        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for key, module in args[0].items():
                self._add_run_module(key, module)
        else:
            for idx, module in enumerate(args):
                self._add_run_module(str(idx), module)

    def _add_run_module(self, name: str, module: Optional[Module]):
        """Add a module to the forward graph of model.

        Args:
            name (str): The name of the module.
            module (nn.Module): The module to add.
        """
        self.add_module(name, module)
        self._runtime_module_list[name] = module
        return self

    def _get_item_by_idx(self, iterator, idx) -> T:
        """Get the idx-th item of the iterator"""
        size = len(self)
        idx = operator.index(idx)
        if not -size <= idx < size:
            raise IndexError('index {} is out of range'.format(idx))
        idx %= size
        return next(islice(iterator, idx, None))

    def __getitem__(self, idx) -> Union[Sequential, T]:
        if isinstance(idx, slice):
            return self.__class__(OrderedDict(list(self._runtime_module_list.items())[idx]))
        else:
            return self._get_item_by_idx(self._runtime_module_list.values(), idx)

    def __setitem__(self, idx: int, module: nn.Module) -> None:
        key: str = self._get_item_by_idx(self._runtime_module_list.keys(), idx)
        return setattr(self, key, module)

    def __delitem__(self, idx: Union[slice, int]) -> None:
        if isinstance(idx, slice):
            for key in list(self._runtime_module_list.keys())[idx]:
                delattr(self, key)
        else:
            key = self._get_item_by_idx(self._runtime_module_list.keys(), idx)
            delattr(self, key)

    def __len__(self) -> int:
        return len(self._runtime_module_list)

    def __dir__(self):
        keys = super(Sequential, self).__dir__()
        keys = [key for key in keys if not key.isdigit()]
        return keys

    def __iter__(self) -> Iterator[nn.Module]:
        return iter(self._runtime_module_list.values())



from __future__ import annotations

import typing

import networkx

from analogvnn.graph.GraphEnum import GraphEnum

if typing.TYPE_CHECKING:
    import graphviz

__all__ = ['to_graphviz_digraph']


def to_graphviz_digraph(from_graph: networkx.DiGraph, real_label: bool = False) -> graphviz.Digraph:
    """Returns a pygraphviz graph from a NetworkX graph N.

    Args:
        from_graph (networkx.DiGraph): the graph to convert.
        real_label (bool): True to use the real label.

    Returns:
        graphviz.Digraph: the converted graph.

    Raises:
        ImportError: if pygraphviz (https://pygraphviz.github.io/) is not available.
        ImportError: if graphviz (https://pygraphviz.github.io/) is not available.
    """
    try:
        import pygraphviz
    except ImportError as e:
        raise ImportError("requires pygraphviz: https://pygraphviz.github.io/") from e
    try:
        from graphviz import Digraph
    except ImportError as e:
        raise ImportError("requires graphviz: https://pygraphviz.github.io/") from e
    strict = networkx.number_of_selfloops(from_graph) == 0 and not from_graph.is_multigraph()
    node_attr = dict(style='filled',
                     shape='box',
                     align='left',
                     fontsize='10',
                     ranksep='0.1',
                     height='0.2',
                     fontname='monospace')
    new_graph = Digraph(
        name=from_graph.name,
        strict=strict,
        node_attr=node_attr,
        graph_attr=dict(size="12,12"),
        format="svg"
    )

    # default graph attributes
    new_graph.graph_attr.update(from_graph.graph.get("graph", {}))
    new_graph.node_attr.update(from_graph.graph.get("node", {}))
    new_graph.edge_attr.update(from_graph.graph.get("edge", {}))

    new_graph.graph_attr.update(
        (k, v) for k, v in from_graph.graph.items() if k not in ("graph", "node", "edge")
    )

    # add nodes
    for n, nodedata in from_graph.nodes(data=True):
        attr = {k: str(v) for k, v in nodedata.items()}
        if isinstance(n, GraphEnum):
            attr["fillcolor"] = 'lightblue'
        attr["name"] = str(id(n))
        attr["label"] = str(n).strip()
        new_graph.node(**attr)

    for u, v, edgedata in from_graph.edges(data=True):
        attr = {k: str(v) for k, v in edgedata.items()}
        attr["tail_name"] = str(id(u))
        attr["head_name"] = str(id(v))

        if real_label and "real_label" in attr:
            attr["label"] = attr["real_label"]
        else:
            attr["label"] = attr["label"].replace("->", "→") if "label" in attr else None
        new_graph.edge(**attr)

    return new_graph



from typing import Optional

import torch
from torch import Tensor

from analogvnn.nn.activation.Activation import Activation

__all__ = ['BinaryStep']


class BinaryStep(Activation):
    """Implements the binary step activation function.
    """

    @staticmethod
    def forward(x: Tensor) -> Tensor:
        """forward pass of the binary step activation function.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return (x >= 0).type(torch.float)

    def backward(self, grad_output: Optional[Tensor]) -> Optional[Tensor]:
        """backward pass of the binary step activation function.

        Args:
            grad_output (Optional[Tensor]): the gradient of the output tensor.

        Returns:
            Optional[Tensor]: the gradient of the input tensor.
        """
        return torch.zeros_like(grad_output)



from typing import Optional

import torch
from torch import Tensor, nn

from analogvnn.nn.activation.Activation import Activation

__all__ = ['Logistic', 'Sigmoid']


class Logistic(Activation):
    """Implements the logistic activation function.
    """

    @staticmethod
    def forward(x: Tensor) -> Tensor:
        """forward pass of the logistic activation function.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return 1 / (1 + torch.exp(-x))

    def backward(self, grad_output: Optional[Tensor]) -> Optional[Tensor]:
        """backward pass of the logistic activation function.

        Args:
            grad_output (Optional[Tensor]): the gradient of the output tensor.

        Returns:
            Optional[Tensor]: the gradient of the input tensor.
        """
        x = self.inputs
        grad = self.forward(x) * (1 - self.forward(x))
        return grad_output * grad

    @staticmethod
    def initialise(tensor: Tensor) -> Tensor:
        """initialize the tensor using xavier uniform initialization with gain associated
        with the logistic activation function.

        Args:
            tensor (Tensor): the tensor to be initialized.

        Returns:
            Tensor: the initialized tensor.
        """
        return nn.init.xavier_uniform(tensor, gain=nn.init.calculate_gain('sigmoid'))

    @staticmethod
    def initialise_(tensor: Tensor) -> Tensor:
        """in-place initialize the tensor using xavier uniform initialization with gain associated
        with the logistic activation function.

        Args:
            tensor (Tensor): the tensor to be initialized.

        Returns:
            Tensor: the initialized tensor.
        """
        return nn.init.xavier_uniform_(tensor, gain=nn.init.calculate_gain('sigmoid'))


class Sigmoid(Logistic):
    """Implements the sigmoid activation function.
    """
    pass



import abc

from torch import Tensor, nn

from analogvnn.backward.BackwardModule import BackwardModule
from analogvnn.nn.module.Layer import Layer

__all__ = ['Activation', 'InitImplement']


class InitImplement:
    """ Implements the initialization of parameters using the activation function.
    """

    @staticmethod
    def initialise(tensor: Tensor) -> Tensor:
        """initialize the tensor using xavier uniform initialization.

        Args:
            tensor (Tensor): the tensor to be initialized.

        Returns:
            Tensor: the initialized tensor.
        """
        return nn.init.xavier_uniform(tensor)

    @staticmethod
    def initialise_(tensor: Tensor) -> Tensor:
        """in-place initialize the tensor using xavier uniform initialization.

        Args:
            tensor (Tensor): the tensor to be initialized.

        Returns:
            Tensor: the initialized tensor.
        """
        return nn.init.xavier_uniform_(tensor)


class Activation(Layer, BackwardModule, InitImplement, abc.ABC):
    """This class is base class for all activation functions.
    """
    pass



from torch import Tensor

__all__ = ['to_matrix']


def to_matrix(tensor: Tensor) -> Tensor:
    """`to_matrix` takes a tensor and returns a matrix with the same values as the tensor.

    Args:
        tensor (Tensor): Tensor

    Returns:
        Tensor: Tensor with the same values as the tensor, but with shape (1, -1).
    """
    if len(tensor.size()) == 1:
        return tensor.view(1, -1)
    return tensor



import math
from typing import Optional

import torch
from torch import Tensor

from analogvnn.nn.activation.Activation import Activation

__all__ = ['Gaussian', 'GeLU']


class Gaussian(Activation):
    """Implements the Gaussian activation function.
    """

    @staticmethod
    def forward(x: Tensor) -> Tensor:
        """forward pass of the Gaussian activation function.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return torch.exp(-torch.pow(x, 2))

    def backward(self, grad_output: Optional[Tensor]) -> Optional[Tensor]:
        """backward pass of the Gaussian activation function.

        Args:
            grad_output (Optional[Tensor]): the gradient of the output tensor.

        Returns:
            Optional[Tensor]: the gradient of the input tensor.
        """
        x = self.inputs
        grad = -2 * x * torch.exp(-torch.pow(x, 2))
        return grad_output * grad


class GeLU(Activation):
    """Implements the Gaussian error linear unit (GeLU) activation function.
    """

    @staticmethod
    def forward(x: Tensor) -> Tensor:
        """forward pass of the Gaussian error linear unit (GeLU) activation function.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor.
        """
        return (1 / 2) * x * (1 + torch.erf(x / math.sqrt(2)))

    def backward(self, grad_output: Optional[Tensor]) -> Optional[Tensor]:
        """backward pass of the Gaussian error linear unit (GeLU) activation function.

        Args:
            grad_output (Optional[Tensor]): the gradient of the output tensor.

        Returns:
            Optional[Tensor]: the gradient of the input tensor.
        """
        x = self.inputs
        grad = (1 / 2) * (
                (1 + torch.erf(x / math.sqrt(2))) + x * ((2 / math.sqrt(math.pi)) * torch.exp(-torch.pow(x, 2)))
        )
        return grad_output * grad



from abc import ABC

from torch import Tensor

from analogvnn.backward.BackwardModule import BackwardModule
from analogvnn.utils.common_types import TENSORS

__all__ = ['BackwardIdentity']


class BackwardIdentity(BackwardModule, ABC):
    """The backward module that returns the output gradients as the input gradients.
    """

    def backward(self, *grad_output: Tensor, **grad_output_kwarg: Tensor) -> TENSORS:
        """Returns the output gradients as the input gradients.

        Args:
            *grad_output (Tensor): The gradients of the output of the layer.
            **grad_output_kwarg (Tensor): The gradients of the output of the layer.

        Returns:
            TENSORS: The gradients of the input of the layer.
        """
        if len(grad_output) == 0:
            return None

        if len(grad_output) == 1:
            return grad_output[0]

        return grad_output
